import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.diabetes_quality_measure import DiabetesQualityMeasure
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    ConditionRecordSet,
    ProtocolOverrideRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class DiabetesQualityMeasureTest(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_abstract(self):
        self.assertTrue(DiabetesQualityMeasure._meta.abstract)

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions = ConditionRecordSet([])
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

        # patient has diabetes --> true
        patient = self.load_patient('diabetes_mixin')
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        patient.billing_line_items = BillingLineItemRecordSet([])
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        protocol.in_initial_population()
        self.assertFalse(protocol.in_initial_population())
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe, context='report')
        self.assertFalse(protocol.in_initial_population())
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe, context='guidance')
        self.assertTrue(protocol.in_initial_population())

        # -- patient to young < 18 --> false
        patient = self.load_patient('diabetes_mixin')
        patient.patient['birthDate'] = '2000-08-24'
        self.assertTrue(17 < patient.age_at(timeframe.end) < 18)
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())
        patient.patient['birthDate'] = '2000-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 18)
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        # -- patient to old > 75 --> false
        patient = self.load_patient('diabetes_mixin')
        patient.patient['birthDate'] = '1943-08-22'
        self.assertTrue(75 < patient.age_at(timeframe.end) < 76)
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())
        patient.patient['birthDate'] = '1943-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 75)
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())
        patient.patient['birthDate'] = '1943-08-24'
        self.assertTrue(patient.age_at(timeframe.end) < 75)
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        # -- patient has an other code --> false
        patient = self.load_patient('diabetes_mixin')
        for condition in patient.conditions:
            for code in condition['coding']:
                code['code'] = 'code'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

        # -- patient has diabetes resolved before the period --> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-22'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

        # -- patient has diabetes resolved within the period --> true
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-24'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        # -- patient has diabetes resolved within the period + active only  --> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-24'
        patient.active_only = True
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

    def test_in_initial_population_with_override(self):
        patient = self.load_patient('diabetes_mixin')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'DiabetesQualityMeasure',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.patient['birthDate'] = '2000-08-24'

        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2018-08-23'))
        self.assertFalse(tested.in_initial_population())

        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2018-08-25'))
        self.assertTrue(tested.in_initial_population())

        patient.billing_line_items = BillingLineItemRecordSet([{
            'id': 92,
            'created': '2075-08-05T21:41:21.407046Z',
            'datetimeOfService': '2075-08-05T21:41:21.407046Z',
            'cpt': '99202',
            'units': 1
        }])
        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2075-08-23'))
        self.assertTrue(tested.in_initial_population())

        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2075-08-25'))
        self.assertFalse(tested.in_initial_population())

    def test_has_diabetes_in_period(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions = ConditionRecordSet([])
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.has_diabetes_in_period())

        # patient has diabetes --> true
        patient = self.load_patient('diabetes_mixin')
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.has_diabetes_in_period())

        # -- patient has diabetes resolved before the period --> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-22'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.has_diabetes_in_period())

        # -- patient has diabetes resolved within the period --> true
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-24'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.has_diabetes_in_period())

        # -- patient has diabetes resolved within the period + active only--> false
        patient = self.load_patient('diabetes_mixin')
        patient.conditions[0]['periods'][0]['to'] = '2017-08-24'
        patient.active_only = True
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.has_diabetes_in_period())

        # -- patient has diabetes after the period --> false
        now = arrow.get('1998-08-22 13:24:56')
        timeframe = Timeframe(start=now, end=now)
        patient = self.load_patient('diabetes_mixin')
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.has_diabetes_in_period())
        now = arrow.get('1998-08-21 13:24:56')
        timeframe = Timeframe(start=now, end=now)
        patient = self.load_patient('diabetes_mixin')
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.has_diabetes_in_period())

    def test_first_due_in(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has diabetes + older than 18 --> none
        patient = self.load_patient('diabetes_mixin')
        patient.patient['birthDate'] = '2000-08-22'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

        # patient has diabetes + younger than 18 --> none
        patient = self.load_patient('diabetes_mixin')
        patient.patient['birthDate'] = '2000-08-30'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertEqual(6, protocol.first_due_in())

        # patient has NO diabetes + younger than 18 --> none
        patient = self.load_patient('diabetes_mixin')
        patient.conditions = ConditionRecordSet([])
        patient.patient['birthDate'] = '2000-08-30'
        protocol = DiabetesQualityMeasure(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

    def test_first_due_in_with_override(self):
        patient = self.load_patient('diabetes_mixin')
        # patient.conditions = ConditionRecordSet([{
        #     'coding': [{
        #         'system': 'ICD-10',
        #         'code': 'E119'
        #     }],
        #     'periods': [{
        #         'from': '2018-05-24',
        #         'to': None
        #     }]
        # }])
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'DiabetesQualityMeasure',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.patient['birthDate'] = '2000-08-24'
        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2018-08-05'))
        result = tested.first_due_in()
        self.assertEqual(19, result)

    def test_specific_visits(self):
        patient = self.load_patient('diabetes_mixin')
        tested = DiabetesQualityMeasure(patient=patient, now=arrow.get('2018-08-05'))
        result = tested.specific_visits
        self.assertIn('cpt', result.values)
        self.assertEqual(24, len(result.values['cpt']))



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    BipolarDisorder, CareServicesInLongTermResidentialFacility, ContactOrOfficeVisit, Dysthymia,
    Ethnicity, MajorDepressionIncludingRemission, OncAdministrativeSex, PalliativeCareEncounter,
    PalliativeOrHospiceCare, Payer, PersonalityDisorderEmotionallyLabile,
    PervasiveDevelopmentalDisorder, Phq9AndPhq9MTools, Race, SchizophreniaOrPsychoticDisorder)


class ClinicalQualityMeasure159v9(ClinicalQualityMeasure):
    """
    Depression Remission at Twelve Months

    Description: The percentage of adolescent patients 12 to 17 years of age and adult patients 18
    years of age or older with major depression or dysthymia who reached remission 12 months (+/-
    60 days) after an index event.

    Definition: Denominator Identification Period:
    The period in which eligible patients can have an index event. The denominator identification
    period occurs prior to the measurement period and is defined as 14 months to two months prior
    to the start of the measurement period. For patients with an index event, there needs to be
    enough time following index for the patients to have the opportunity to reach remission twelve
    months +/- 60 days after the index event date.

    Index Event Date:
    The date in which the first instance of elevated PHQ-9 or PHQ-9M greater than nine and
    diagnosis of depression or dysthymia occurs during the denominator identification measurement
    period. Patients may be screened using PHQ-9 and PHQ-9M up to 7 days prior to the office visit
    (including the day of the office visit).


    Measure Assessment Period:
    The index event date marks the start of the measurement assessment period for each patient
    which is 14 months (12 months +/- 60 days) in length to allow for a follow-up PHQ-9 or PHQ-9M
    between 10 and 14 months following the index event. This assessment period is fixed and does
    not start over with a higher PHQ-9 or PHQ-9M that may occur after the index event date.

    Remission is defined as a PHQ-9 or PHQ-9M score of less than five.

    Twelve months is defined as the point in time from the index event date extending out twelve
    months and then allowing a grace period of sixty days prior to and sixty days after this date.
    The most recent PHQ-9 or PHQ-9M score less than five obtained during this four month period is
    deemed as remission at twelve months, values obtained prior to or after this period are not
    counted as numerator compliant (remission).

    Rationale: Adults:
    Depression is a common and treatable mental disorder. 8.1% of American adults age 20 and over
    had depression in a given 2 week period. Women (10.4%) were almost twice as likely as men
    (5.5%) to have had depression. The prevalence of depression among adults decreased as family
    income levels increased. About 80% of adults with depression reported at least some difficulty
    with work, home, or social activities because of their depression symptoms (Centers for Disease
    Control and Prevention, 2018).

    Depression is a risk factor for development of chronic illnesses such as diabetes and CHD and
    adversely affects the course, complications and management of chronic medical illness. Both
    maladaptive health risk behaviors and psychobiological factors associated with depression may
    explain depression's negative effect on outcomes of chronic illness. (Katon, W.J., 2011)

    Adolescents and Adults:
    The Centers for Disease Control and Prevention states that during 2009-2012 an estimated 7.6%
    of the U.S. population aged 12 and over had depression, including 3% of Americans with severe
    depressive symptoms. Almost 43% of persons with severe depressive symptoms reported serious
    difficulties in work, home and social activities, yet only 35% reported having contact with a
    mental health professional in the past year.
    Depression is associated with higher mortality rates in all age groups. People who are
    depressed are 30 times more likely to take their own lives than people who are not depressed
    and five times more likely to abuse drugs. Depression is the leading cause of medical
    disability for people aged 14 to 44. Depressed people lose 5.6 hours of productive work every
    week when they are depressed, fifty percent of which is due to absenteeism and short-term
    disability.

    Adolescents:
    In 2014, an estimated 2.8 million adolescents age 12 to 17 in the United States had at least
    one major depressive episode (MDE) in the past year. This represented 11.4% of the U.S.
    population. The same survey found that only 41.2 percent of those who had a MDE received
    treatment in the past year. The 2013 Youth Risk Behavior Survey of students grades 9 to 12
    indicated that during the past 12 months 39.1% (F) and 20.8% (M) indicated feeling sad or
    hopeless almost every day for at least 2 weeks, planned suicide attempt 16.9% (F) and 10.3%
    (M), with attempted suicide 10.6% (F) and 5.4% (M). Adolescent-onset depression is associated
    with chronic depression in adulthood. Many mental health conditions (anxiety, bipolar,
    depression, eating disorders, and substance abuse) are evident by age 14. The 12-month
    prevalence of MDEs increased from 8.7% in 2005 to 11.3% in 2014 in adolescents and from 8.8% to
    9.6% in young adults (both P < .001). The increase was larger and statistically significant
    only in the age range of 12 to 20 years. The trends remained significant after adjustment for
    substance use disorders and sociodemographic factors. Mental health care contacts overall did
    not change over time; however, the use of specialty mental health providers increased in
    adolescents and young adults, and the use of prescription medications and inpatient
    hospitalizations increased in adolescents. In 2015, 9.7% of adolescents in MN who were screened
    for depression or other mental health conditions, screened positively.

    Guidance: When a baseline assessment is conducted with PHQ 9M, the follow-up assessment can use
    either a PHQ 9M or PHQ 9.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms159v9
    """

    title = 'Depression Remission at Twelve Months'

    identifiers = ['CMS159v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'MN Community Measurement',
    ]

    references = [
        'Duffy, F. F., Chung, H., Trivedi, M., et al. (2008, October). Systematic use of patient-rated depression severity monitoring: Is it helpful and feasible in clinical psychiatry? Psychiatric Services, 59(10), 1148-1154.',
        'Hunkeler, E. M., Meresman, J. F., Hargreaves, W. A., et al. (2000, August). Efficacy of nurse telehealth care and peer support in augmenting treatment of depression in primary care. Archives of Family Medicine, 9(8), 700-708.',
        'Katon, W.J. (2011) Epidemiology and treatment of depression in patients with chronic medical illness Dialogues in Clinical Neuroscience v.13(1) PMC3181964',
        'Kroenke, K., Spitzer, R. L., & Williams, J. B. W. (2001). The PHQ-9: Validity of a brief depression severity measure. Journal of General Internal Medicine, 16(9), 606-613. Retrieved from https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1495268/',
        'Lowe, B., Unutzer, J., Callahan, C. M., et al. (2004). Monitoring depression treatment outcomes with the Patient Health Questionnaire-9. Medical Care, 42(12), 1194-1201.',
        'Rush, A. J., Trivedi, M. H., Wisniewski, S. R., et al. (2006). Acute and longer-term outcomes in depressed outpatients requiring one or several treatment steps: A STAR*D report. American Journal of Psychiatry, 163, 1905-1917.',
        'Simon, G. E., Van Korff, M., Rutter, C., et al. (2000). Randomised trial of monitoring, feedback, and management of care by telephone to improve treatment of depression in primary care. BMJ, 320, 550-554.',
        'Trangle , M., Gursky, J., Haight, R., et al. Depression, adults in primary care. (Updated 2016, March). Retrieved from https://www.icsi.org/guidelines__more/catalog_guidelines_and_more/catalog_guidelines/catalog_behavioral_health_guidelines/depression/',
        'Trivedi, M. H. (2009). Tools and strategies for ongoing assessment of depression: A measurement-based approach to remission. Journal of Clinical Psychiatry, 70, 26-31.',
        'Trivedi, M. H., Rush, A. J., Wisniewski, S. R., et al. (2006). Evaluation of outcomes with citalopram for depression using measurement-based care in STAR*D: Implications for clinical practice. American Journal of Psychiatry, 163(1), 28-40.',
        'Unutzer, J., Katon, W., Callahan, C. M., et al. (2002). Collaborative care management of late-life depression in the primary care setting: A randomized controlled trial. JAMA, 288, 2836-2845.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Adolescent patients 12 to 17 years of age and adult patients 18 years
        of age and older with a diagnosis of major depression or dysthymia and an initial PHQ-9 or
        PHQ-9M score greater than nine during the index event. Patients may be screened using PHQ-9
        and PHQ-9M up to 7 days prior to the office visit (including the day of the office visit).
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: 1: Patients who died
        2: Patients who received hospice or palliative care services
        3: Patients who were permanent nursing home residents
        4: Patients with a diagnosis of bipolar disorder
        5: Patients with a diagnosis of personality disorder emotionally labile
        6: Patients with a diagnosis of schizophrenia or psychotic disorder
        7: Patients with a diagnosis of pervasive developmental disorder

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Adolescent patients 12 to 17 years of age and adult patients 18 years of age and
        older who achieved remission at twelve months as demonstrated by a twelve month (+/- 60
        days) PHQ-9 or PHQ-9M score of less than five

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Adults:
        Source: Institute for Clinical Systems Improvement (ICSI) Health Care Guideline for Adult
        Depression in Primary Care (Trangle et al., 2016)

        Recommendations and algorithm notations supporting depression outcomes and duration of
        treatment according to ICSI's Health Care Guideline:

        Recommendation: Clinicians should establish and maintain follow-up with patients.
        Appropriate, reliable follow-up is highly correlated with improved response and remission
        scores. It is also correlated with the improved safety and efficacy of medications and
        helps prevent relapse.

        Proactive follow-up contacts (in person, telephone) based on the collaborative care model
        have been shown to significantly lower depression severity (Unutzer et al., 2002).  In the
        available clinical effectiveness trials conducted in real clinical practice settings, even
        the addition of a care manager leads to modest remission rates (Trivedi et al., 2006;
        Unutzer et al., 2002).  Interventions are critical to educating the patient regarding the
        importance of preventing relapse, safety and efficacy of medications, and management of
        potential side effects.  Establish and maintain initial follow-up contact intervals
        (office, phone, other) (Hunkeler et al., 2000; Simon et al., 2000).

        PHQ-9 as monitor and management tool.  The PHQ-9 is an effective management tool, as well,
        and should be used routinely for subsequent visits to monitor treatment outcomes and
        severity. It can also help the clinician decide if/how to modify the treatment plan (Duffy
        et al., 2008; Lowe et al., 2004).  Using a measurement-based approach to depression care,
        PHQ-9 results and side effect evaluation should be combined with treatment algorithms to
        drive patients toward remission.  A five-point drop in PHQ-9 score is considered the
        minimal clinically significant difference (Trivedi, 2009).

        The goals of treatment should be to achieve remission, reduce relapse and recurrence, and
        return to previous level of occupational and psychosocial function.

        If using a PHQ-9 tool, remission translates to PHQ-9 score of less than 5 (Kroenke, 2001).
        Results from the STAR*D study showed that remission rates lowered with more treatment
        steps, but the overall cumulative rate was 67% (Rush et al., 2006).

        Response and remission take time. In the STAR*D study, longer times than expected were
        needed to reach response or remission. In fact, one-third of those who ultimately responded
        did so after six weeks. Of those who achieved remission by Quick Inventory of Depressive
        Symptomatology (QIDS), 50% did so only at or after six weeks of treatment (Trivedi et al.,
        2006). If the primary care clinician is seeing some improvement, continue working with that
        patient to augment or increase dosage to reach remission. This can take up to three months.

        This measure assesses achievement of remission, which is a desired outcome of effective
        depression treatment and monitoring.

        Adult Depression in Primary Care - Guideline Aims
        - Increase the percentage of patients with major depression or persistent depressive
        disorder who have improvement in outcomes from treatment for major depression or persistent
        depressive disorder.
        - Increase the percentage of patients with major depression or persistent depressive
        disorder who have follow-up to assess for outcomes from treatment.
        - Improve communication between the primary care physician and the mental health care
        clinician (if patient is co-managed).

        Adolescents:
        Source: American Academy of Child and Adolescent Psychiatry Practice Parameter for the
        Assessment and Treatment of Children and Adolescents with Depressive Disorders (2007)
        http://www.jaacap.com/article/S0890-8567(09)62053-0/pdf

        Recommendations:
        Recommendations supporting depression outcomes and duration of treatment according to AACAP
        guideline:
        - Treatment of depressive disorders should always include an acute and continuation phase;
        some children may also require maintenance treatment. The main goal of the acute phase is
        to achieve response and ultimately full symptomatic remission (definitions below).
        - Each phase of treatment should include psychoeducation, supportive management, and family
        and school involvement.
        - Education, support, and case management appear to be sufficient treatment for the
        management of depressed children and adolescents with an uncomplicated or brief depression
        or with mild psychosocial impairment.
        - For children and adolescents who do not respond to supportive psychotherapy or who have
        more complicated depressions, a trial with specific types of psychotherapy and/or
        antidepressants is indicated.

        Sources:
        Guidelines for Adolescent Depression in Primary Care (GLAD-PC) (2018)
        http://pediatrics.aappublications.org/content/141/3/e20174081
        Guidelines for adolescent depression in primary care (GLAD-PC): II. Treatment and ongoing
        management
        http://pediatrics.aappublications.org/content/141/3/e20174082

        Recommendations supporting depression outcomes and duration of treatment according to GLAD-
        PC:
        Recommendations for Ongoing Management of Depression:
        - Mild depression: consider a period of active support and monitoring before starting other
        evidence-based treatment
        - Moderate or severe major clinical depression or complicating factors:
          -- consultation with mental health specialist with agreed upon roles
          -- evidence based treatment (CBT or IPT and/or antidepressant SSRI)
        - Monitor for adverse effects during antidepressant therapy
          -- clinical worsening, suicidality, unusual changes in behavior
        - Systematic and regular tracking of goals and outcomes
          -- improvement in functioning status and resolution of depressive symptoms
        Regardless of the length of treatment, all patients should be monitored on a monthly basis
        for 6 to 12 months after the full resolution of symptoms
        """
        pass



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    Ethnicity, HumanImmunodeficiencyVirusHivLaboratoryTestCodesAbAndAg,
    IndicatorsOfHumanImmunodeficiencyVirusHiv, OfficeVisit, OncAdministrativeSex, Payer,
    PreventiveCareEstablishedOfficeVisit0To17, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit0To17, PreventiveCareServicesInitialOfficeVisit18AndUp,
    Race)


class ClinicalQualityMeasure349v3(ClinicalQualityMeasure):
    """
    HIV Screening

    Description: Percentage of patients aged 15-65 at the start of the measurement period who were
    between 15-65 years old when tested for HIV

    Definition: None

    Rationale: Human immunodeficiency virus (HIV) is a communicable infection that leads to a
    progressive disease with a long asymptomatic period. There were an estimated 38,700 new HIV
    infections in the United States in 2016 (Centers for Disease Control and Prevention, 2019a).
    Without treatment, most persons develop acquired immunodeficiency syndrome (AIDS) within 10
    years of HIV infection. Antiretroviral therapy (ART) delays this progression and increases the
    length of survival, but it is most effective when initiated during the asymptomatic phase.
    Persons living with HIV who use ART and achieve viral suppression can have a nearly normal life
    expectancy (Samji et al., 2013).  DHHS Guidelines for the Use of Antiretroviral Agents in
    HIV-1-Infected Adults and Adolescents recommends antiretroviral therapy for all HIV-infected
    individuals to reduce the risk of disease progression (regardless of CD4 cell count at
    diagnosis) (Panel on Antiretroviral Guidelines for Adults and Adolescents, 2017).

    CDC estimates that, at the end of 2016, approximately 14% of the 1.1 million adults and
    adolescents living with HIV infection in the United States were unaware of their infection
    (Centers for Disease Control and Prevention, 2018b). Among persons diagnosed with HIV in 2017,
    approximately 21% were diagnosed with Stage 3 HIV (AIDS) at the time of HIV diagnosis (Centers
    for Disease Control and Prevention, 2019c), which is when the median CD4 count at diagnosis is
    less than 200 cells/mm3 for persons aged greater than or equal to 6 years (Centers for Disease
    Control and Prevention, 2019a). HIV screening identifies infected persons who were previously
    unaware of their infection, which enables them to seek medical and social services that can
    improve their health and the quality and length of their lives. Additionally, using ART with
    high levels of medication adherence has been shown to substantially reduce risk for HIV
    transmission (Panel on Antiretroviral Guidelines for Adults and Adolescents, 2017).

    Based on the Behavioral Risk Factor Surveillance System (BRFSS), the percentage of ever tested
    for HIV increased from 42.9% in 2011 to 45.9% in 2017. Despite this increase, less than half of
    US adults have ever been tested for HIV over ten years after CDC’s recommendations (Patel et
    al., 2019).

    Guidance: This measure evaluates the proportion of patients aged 15 to 65 at the start of the
    measurement period who have documentation of having received an HIV test at least once on or
    after their 15th birthday and before their 66th birthday. In order to satisfy the measure, the
    reporting provider must have documentation of the administration of the laboratory test present
    in the patient's medical record. In cases where the HIV test was performed elsewhere, providers
    cannot rely on patient attestation or self-report to meet the measure requirements, as previous
    research has shown that patient self-report is an unreliable indicator of previous HIV testing
    history. Rather, providers must request documentation of those test results. If such
    documentation is not available, the patient should be considered still eligible for HIV
    screening. If such documentation is available, but cannot be provided in a standardized,
    structured format (such that the lab test and results can be readily incorporated as structured
    data within the EHR), providers should enter the information into their EHR as a laboratory
    test in a manner consistent with the EHR in use. If the specific Human Immunodeficiency Virus
    (HIV) Laboratory Test LOINC code of the test is not known, the entry should use the more
    generic code LOINC panel code [75622-1].

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms349v3
    """

    title = 'HIV Screening'

    identifiers = ['CMS349v3']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'Mathematica',
    ]

    references = [
        'Centers for Disease Control and Prevention. (2019a). Estimated HIV incidence and prevalence in the United States, 2010–2016; Table 1, page 20. HIV Surveillance Supplemental Report, 24(1). Retrieved from http://www.cdc.gov/hiv/library/reports/hiv-surveillance.html',
        'Centers for Disease Control and Prevention. (2019b). Estimated HIV incidence and prevalence in the United States, 2010–2016. HIV Surveillance Supplemental Report, 24(1). Retrieved from https://www.cdc.gov/hiv/pdf/library/reports/surveillance/cdc-hiv-surveillance-supplemental-report-vol-24-1.pdf',
        'Centers for Disease Control and Prevention. (2019c, June). Monitoring selected national HIV prevention and care objectives by using HIV surveillance data—United States and 6 dependent areas, 2017. HIV Surveillance Supplemental Report, 24(3). Retrieved from https://www.cdc.gov/hiv/pdf/library/reports/surveillance/cdc-hiv-surveillance-supplemental-report-vol-24-3.pdf',
        'Centers for Disease Control and Prevention. (2006). Revised recommendations for HIV testing of adults, adolescents, and pregnant women in health care settings. Morbidity and Mortality Weekly Report, 55(RR-14), 1-17.',
        'Owens, on behalf of the U.S. Preventive Task Force. (2019). Screening for HIV: US Preventive Services Task Force Recommendation Statement. JAMA, 321(23), 2326-2336.',
        'Panel on Antiretroviral Guidelines for Adults and Adolescents. (2017). Guidelines for the use of antiretroviral agents in HIV-1-infected adults and adolescents. Retrieved from https://aidsinfo.nih.gov/contentfiles/lvguidelines/adultandadolescentgl.pdf',
        'Patel, D., Johnson, C.H., Krueger, A. et al. (2019). Trends in HIV Testing Among US Adults, Aged 18–64 Years, 2011–2017. AIDS Behav.',
        'Samji, H., Cescon, A., Hogg, R. S., et al. (2013). Closing the gap: Increases in life expectancy among treated HIV-positive individuals in the United States and Canada. PLoS One, 8, e81355.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 15 to 65 years of age at the start of the measurement period
        AND who had at least one outpatient visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients diagnosed with HIV prior to the start of the measurement period

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with documentation of an HIV test performed on or after their 15th
        birthday and before their 66th birthday

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The US Preventive Services Task Force recommends that clinicians
        screen for HIV infection in adolescents and adults aged 15 to 65 years. Younger adolescents
        and older adults who are at increased risk should also be screened (A Recommendation)
        (Owens, et al., 2019).

        Since 2006, the CDC has recommended routine opt-out HIV screening (i.e., patient is
        notified that testing will be performed unless the patient declines) in healthcare
        facilities of adolescents and adults 13-64 years of age and HIV diagnostic testing of
        adolescents and adults with clinical signs or symptoms consistent with HIV infection
        (Centers for Disease Control and Prevention, 2006).
        """
        pass



# type: ignore
from canvas_workflow_kit.builtin_cqms.cms138v6_preventive_care_and_screening_tobacco_use_screening_and_cessation_intervention import (  # noqa: E501
    STATUS_DUE,
    STATUS_NOT_APPLICABLE,
    ClinicalQualityMeasure138v6,
    InstructionRecommendation,
    PrescribeRecommendation,
    ProtocolResult,
    TobaccoUseCessationCounseling,
    TobaccoUseCessationPharmacotherapy,
    events
)


class ClinicalQualityMeasure138v6p2(ClinicalQualityMeasure138v6):
    """
    Use only the Population 2's path of the ClinicalQualityMeasure138v6
    """

    class Meta:
        title = 'Preventive Care and Screening: Tobacco Use: Cessation Intervention'
        version = '2019-04-18v1'

        description = ('Patients aged 18 years and older, and identified as a tobacco user, '
                       'who have not received tobacco cessation intervention such as counselling, '
                       'referral or medication.')
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS138v6.html'

        identifiers = ['CMS138v6p2']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'American Medical Association (AMA)',
            'PCPI(R) Foundation (PCPI[R])',
        ]

    def in_initial_population(self) -> bool:
        super().in_initial_population()
        return self._populations[self.POPULATION_2].in_initial_population

    def in_denominator(self) -> bool:
        super().in_denominator()
        return self._populations[self.POPULATION_2].in_denominator

    def in_numerator(self) -> bool:
        super().in_numerator()
        return self._populations[self.POPULATION_2].in_numerator

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            self.in_numerator()
            if (self._populations[self.POPULATION_2].in_denominator and
                    not self._populations[self.POPULATION_2].in_numerator):
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(f'{self.patient.first_name} is a current tobacco user, '
                                     f'intervention is indicated.')
                result.add_recommendation(
                    InstructionRecommendation(
                        key='CMS138v6p2_RECOMMEND_CESSATION_COUNSELING',
                        rank=1,
                        button='Plan',
                        patient=self.patient,
                        instruction=TobaccoUseCessationCounseling,
                        title='Tobacco cessation counseling'))
                result.add_recommendation(
                    PrescribeRecommendation(
                        key='CMS138v6p2_RECOMMEND_CESSATION_MEDICATION',
                        rank=2,
                        button='Plan',
                        patient=self.patient,
                        prescription=TobaccoUseCessationPharmacotherapy,
                        title='Cessation support medication'))
            elif self.tobacco_cessation_intervention_counseling:
                self.satisfied_result(self.tobacco_cessation_intervention_counseling,
                                      '{name} had a smoking cessation counseling {date}.', result)
            elif self.tobacco_cessation_intervention_medication:
                self.satisfied_result(self.tobacco_cessation_intervention_medication,
                                      '{name} has been prescribed cessation medication {date}.',
                                      result)
        elif self.patient.age_at(self.timeframe.end) < self.MINIMUM_AGE:
            result.status = STATUS_NOT_APPLICABLE
            result.due_in = (
                self.patient.birthday.shift(years=self.MINIMUM_AGE) - self.timeframe.end).days
        return result



from canvas_workflow_kit.constants import CHANGE_TYPE
from canvas_workflow_kit.utils import parse_class_from_python_source
from .base import SDKBaseTest
from canvas_workflow_kit import settings
from canvas_workflow_kit import events
from pathlib import Path
from canvas_workflow_kit.protocol import (
    ProtocolResult,
    STATUS_DUE,
    STATUS_SATISFIED,
    STATUS_NOT_APPLICABLE
)


class TemplateTest(SDKBaseTest):

    def setUp(self):
        super().setUp()
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

        patient = self.load_patient('patient')
        self.base_class = self.createProtocolClass({})(patient=patient)
        self.satisfied_class = self.createProtocolClass(
            {'denominator': 'return True', 'numerator': 'return True'})(
            patient=patient)
        self.due_class = self.createProtocolClass({'denominator': 'return True',
                                                   'numerator': 'return False'})(patient=patient)

    def createProtocolClass(self, fields={}):
        template_fields = {
            'value_sets': '*',
            'year': '2022',
            'class_name': "MyClinicalQualityMeasure",
            'title': "My Clinical Quality Measure",
            'description': 'Description',
            'information_link': 'https://link_to_protocol_information',
            'types': ['CQM'],
            'authors': '"Canvas Example Medical Association (CEMA)"',
            'references': '"Gunderson, Beau;  Excellence in Software Templates. 2012 Oct 20;161(7):422-34."',
            'funding_source': '',
            'denominator': 'pass',
            'numerator': 'pass',
            **fields
        }
        template_path = Path(__file__).parent.parent / 'builtin_cqms/stub_template_user.py.txt'
        template = template_path.open('r').read()

        content = template.format(**template_fields)
        return parse_class_from_python_source(content)

    def test_base_fields(self):
        Protocol = self.base_class
        self.assertEqual('Description', Protocol._meta.description)
        self.assertEqual('My Clinical Quality Measure', Protocol._meta.title)
        self.assertEqual('2022-v1', Protocol._meta.version)
        self.assertEqual('https://link_to_protocol_information', Protocol._meta.information)
        self.assertEqual(['CMSMyClinicalQualityMeasure'], Protocol._meta.identifiers)
        self.assertEqual(['CQM'], Protocol._meta.types)
        self.assertEqual([events.HEALTH_MAINTENANCE], Protocol._meta.responds_to_event_types)
        self.assertEqual([CHANGE_TYPE.CONDITION], Protocol._meta.compute_on_change_types)
        self.assertEqual(['Canvas Example Medical Association (CEMA)'], Protocol._meta.authors)
        self.assertEqual(
            ['Gunderson, Beau;  Excellence in Software Templates. 2012 Oct 20;161(7):422-34.'],
            Protocol._meta.references)
        self.assertEqual('', Protocol._meta.funding_source)

    def test_base_numerator(self):
        tested = self.base_class

        numerator = tested.in_numerator()
        self.assertIsNone(numerator)

    def test_base_denominator(self):
        tested = self.base_class

        denominator = tested.in_denominator()
        self.assertIsNone(denominator)

    def test_base_result(self):
        tested = self.base_class
        result = tested.compute_results()

        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual(STATUS_NOT_APPLICABLE, result.status)
        self.assertEqual([], result.recommendations)
        self.assertEqual('', result.narrative)
        self.assertIsNone(result.due_in)
        self.assertEqual(30, result.days_of_notice)
        self.assertIsNone(result.next_review)

    def test_satisfied_numerator(self):
        tested = self.satisfied_class

        numerator = tested.in_numerator()
        self.assertTrue(numerator)

    def test_satisfied_denominator(self):
        tested = self.satisfied_class

        denominator = tested.in_denominator()
        self.assertTrue(denominator)

    def test_satisfied_result(self):
        tested = self.satisfied_class

        result = tested.compute_results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual(STATUS_SATISFIED, result.status)
        self.assertEqual([], result.recommendations)
        self.assertEqual('', result.narrative)
        self.assertIsNone(result.due_in)
        self.assertEqual(30, result.days_of_notice)
        self.assertIsNone(result.next_review)

    def test_due_numerator(self):
        tested = self.due_class

        numerator = tested.in_numerator()
        self.assertFalse(numerator)

    def test_due_denominator(self):
        tested = self.due_class

        denominator = tested.in_denominator()
        self.assertTrue(denominator)

    def test_due_result(self):
        tested = self.due_class

        result = tested.compute_results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual(STATUS_DUE, result.status)
        self.assertEqual(
            'Nicolas has at least two eGFR measurements < 60 ml/min over the last two years suggesting renal disease. There is no associated condition on the Conditions List.',
            result.narrative)
        self.assertEqual(-1, result.due_in)
        self.assertEqual(30, result.days_of_notice)
        self.assertIsNone(result.next_review)
        self.assertEqual(len(result.recommendations), 1)

    def test_due_result_recommendation(self):
        tested = self.due_class

        result = tested.compute_results()
        recommendation = result.recommendations[0]
        self.assertEqual(recommendation.key, 'HCC002v2_RECOMMEND_DIAGNOSE')
        self.assertEqual(recommendation.rank, 1)
        self.assertEqual(recommendation.button, 'Diagnose')
        self.assertEqual(
            recommendation.title,
            'Consider updating the Conditions List to include kidney related problems as clinically appropriate')
        self.assertEqual(
            recommendation.narrative,
            'Nicolas has at least two eGFR measurements < 60 ml/min over the last two years suggesting renal disease. There is no associated condition on the Conditions List.')
        self.assertEqual(recommendation.command, {'type': 'diagnose'})



# type: ignore
from typing import Optional

from canvas_workflow_kit.protocol import CONTEXT_REPORT, ClinicalQualityMeasure, ProtocolResult
from canvas_workflow_kit.value_set.v2018 import (
    AnnualWellnessVisit,
    Diabetes,
    HomeHealthcareServices,
    OfficeVisit,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp
)
from canvas_workflow_kit.value_set.value_set import ValueSet


class DiabetesQualityMeasure(ClinicalQualityMeasure):

    class Meta:
        title = 'Diabetes Quality Measure'
        version = '1.0'
        changelog = 'Initial release'

        description = ''
        information = ''

        identifiers = []

        types = ['CCP']

        authors = [
            'Canvas Medical Team',
        ]

        abstract = True

    AGE_RANGE_START = 18
    AGE_RANGE_END = 75

    @classmethod
    def enabled(cls) -> bool:
        # necessary to prevent error based on not implementing the methods
        return True

    def in_initial_population(self) -> bool:
        """
        Initial population: Patients 18-75 years of age with diabetes with a visit during the
        measurement period
        """
        return (self.patient.age_at_between(self.timeframe.end, self.AGE_RANGE_START,
                                            self.AGE_RANGE_END) and
                self.has_diabetes_in_period() and
                (self.patient.has_visit_within(self.timeframe, self.specific_visits)
                 if self.context == CONTEXT_REPORT else True))

    def has_diabetes_in_period(self):
        if self.patient.conditions.find(Diabetes).intersects(
                self.timeframe, still_active=self.patient.active_only):
            return True
        return False

    def first_due_in(self) -> Optional[int]:
        if (self.patient.age_at(self.timeframe.end) < self.AGE_RANGE_START and
                self.has_diabetes_in_period()):
            return (
                self.patient.birthday.shift(years=self.AGE_RANGE_START) - self.timeframe.end).days
        return None

    def in_denominator(self) -> bool:
        raise NotImplementedError('in_denominator must be overridden')

    def in_numerator(self) -> bool:
        raise NotImplementedError('in_numerator must be overridden')

    def compute_results(self) -> ProtocolResult:
        raise NotImplementedError('compute_results must be overridden')

    @property
    def specific_visits(self) -> ValueSet:
        return (
            OfficeVisit |
            # notes will be enriched with encounter logic to manage
            # FaceToFaceInteraction codings (Andrew, 2019-03-11)
            # FaceToFaceInteraction (SNOMED)
            PreventiveCareServicesEstablishedOfficeVisit18AndUp |
            PreventiveCareServicesInitialOfficeVisit18AndUp | HomeHealthcareServices |
            AnnualWellnessVisit)



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AmitriptylineHydrochloride, Amobarbital, Amoxapine, AnnualWellnessVisit, AntiInfectivesOther,
    Atropine, Benztropine, Brompheniramine, Butabarbital, Butalbital, Carbinoxamine,
    CareServicesInLongTermResidentialFacility, Carisoprodol, Chlorpheniramine, Chlorpropamide,
    Chlorzoxazone, Clemastine, Clomipramine, ConjugatedEstrogens, CyclobenzaprineHydrochloride,
    Cyproheptadine, DesiccatedThyroid, Desipramine, Dexbrompheniramine, Dexchlorpheniramine,
    Dicyclomine, Dimenhydrinate, Diphenhydramine, DiphenhydramineHydrochloride, Dipyridamole,
    DischargeServicesNursingFacility, Disopyramide, Doxylamine, EncounterInpatient,
    ErgoloidMesylates, EsterifiedEstrogens, Estradiol, Estropipate, Ethnicity, Glyburide,
    Guanfacine, HomeHealthcareServices, HospiceCareAmbulatory, Hydroxyzine, Hyoscyamine,
    Imipramine, Indomethacin, Isoxsuprine, KetorolacTromethamine,
    ListOfSingleRxnormCodeConceptsForHighRiskDrugsForTheElderly, Meclizine, Megestrol, Meperidine,
    Meprobamate, Metaxalone, Methocarbamol, Methyldopa, Nifedipine, NonbenzodiazepineHypnotics,
    Nortriptyline, NursingFacilityVisit, OfficeVisit, OncAdministrativeSex, OphthalmologicServices,
    Orphenadrine, Paroxetine, Payer, Pentobarbital, Phenobarbital,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, PromethazineHydrochloride, Propantheline,
    Protriptyline, Race, Scopolamine, Secobarbital, Trihexyphenidyl, Trimipramine, Triprolidine)


class ClinicalQualityMeasure156v9(ClinicalQualityMeasure):
    """
    Use of High-Risk Medications in Older Adults

    Description: Percentage of patients 65 years of age and older who were ordered at least two of
    the same high-risk medications.

    Definition: A high-risk medication is identified by either of the following:
         a. A prescription for medications classified as high risk at any dose and for any duration
         b. Prescriptions for medications classified as high risk at any dose with greater than a
    90 day supply

    Rationale: Certain medications (MacKinnon & Hepler, 2003) are associated with increased risk of
    harm from drug side-effects and drug toxicity and pose a concern for patient safety. There is
    clinical consensus that these drugs pose increased risks in older adults (Kaufman, Brodin, &
    Sarafian, 2005). Potentially inappropriate medication use in older adults has been connected to
    significantly longer hospital stay lengths and increased hospitalization costs (Hagstrom et
    al., 2015) as well as increased risk of death (Lau et al. 2004).

    Older adults receiving inappropriate medications are more likely to report poorer health status
    at follow-up, compared to those who receive appropriate medications (Fu, Liu, & Christensen,
    2004). A study of the prevalence of potentially inappropriate medication use in older adults
    found that 40 percent of individuals 65 and older filled at least one prescription for a
    potentially inappropriate medication and 13 percent filled two or more (Fick et al., 2008).
    While some adverse drug events are unavoidable, studies estimate that between 30 and 80 percent
    of adverse drug events in older adults are preventable (MacKinnon & Hepler, 2003).

    Reducing the number of inappropriate prescriptions can lead to improved patient safety and
    significant cost savings.  Conservative estimates of extra costs due to potentially
    inappropriate medications in older adults average $7.2 billion a year (Fu et al., 2007 ).
    Medication use by older adults will likely increase further as the U.S. population ages, new
    drugs are developed, and new therapeutic and preventive uses for medications are discovered
    (Rothberg et al., 2008). The annual direct costs of preventable adverse drug events (ADEs) in
    the Medicare population have been estimated to exceed $800 million (Institute of Medicine,
    2007). By the year 2030, nearly one in five U.S. residents is expected to be aged 65 years or
    older; this age group is projected to more than double in number from 38.7 million in 2008 to
    more than 88.5 million in 2050.  Likewise, the population aged 85 years or older is expected to
    increase almost four-fold, from 5.4 million to 19 million between 2008 and 2050.  As the older
    adult population continues to grow, the number of older adults who present with multiple
    medical conditions for which several medications are prescribed will continue to increase,
    resulting in polypharmacy concerns (Gray & Gardner, 2009).

    Guidance: The intent of the measure is to assess if the patient has been prescribed at least
    two of the same high-risk medications on different days.

    The intent of the measure is to assess if the reporting provider ordered the high-risk
    medication(s). If the patient had a high-risk medication previously prescribed by another
    provider, they would not be counted towards the numerator unless the reporting provider also
    ordered a high-risk medication for them.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms156v9
    """

    title = 'Use of High-Risk Medications in Older Adults'

    identifiers = ['CMS156v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Geriatrics Society 2015 Beers Criteria Update Expert Panel. (2015). American Geriatrics Society 2015 updated Beers criteria for potentially inappropriate medication use in older adults. Journal of the American Geriatrics Society, 63(11), 2227-2246.',
        'Beers, M. H. (1997). Explicit criteria for determining potentially inappropriate medication use by the elderly. Archives of Internal Medicine, 157, 1531-1536.',
        'Campanelli, C. M. (2012). American Geriatrics Society updated Beers criteria for potentially inappropriate medication use in older adults: The American Geriatrics Society 2012 Beers Criteria Update Expert Panel. Journal of the American Geriatrics Society, 60(4), 616.',
        'Fick, D. M., Cooper J. W., Wade, W. E., et al. (2003). Updating the Beers criteria for potentially inappropriate medication use in older adults. Archives of Internal Medicine, 163(22), 2716-2724.',
        'Fick, D. M., Mion, L. C., Beers, M. H., et al. (2008). Health outcomes associated with potentially inappropriate medication use in older adults. Research in Nursing & Health, 31(1), 42-51.',
        'Fu, A. Z., Liu, G. G., & Christensen, D. B. (2004). Inappropriate medication use and health outcomes in the elderly. Journal of the American Geriatrics Society, 52(11), 1934-1939.',
        'Gray, C. L., & Gardner, C. (2009). Adverse drug events in the elderly: An ongoing problem. Journal of Managed Care & Specialty Pharmacy, 15(7), 568-571.',
        'Hagstrom, K., Nailor, M., Lindberg, M., Hobbs, L., & Sobieraj, D. M. 2015. Association Between Potentially Inappropriate Medication Use in Elderly Adults and Hospital-Related Outcomes. Journal of the American Geriatrics Society, 63(1), 185-186.',
        'Institute of Medicine, Committee on Identifying and Preventing Medication Errors. (2007). Preventing medication errors. Aspden, P., Wolcott, J. A., Bootman, J. L., & Cronenwatt, L. R. (eds.). Washington, DC: National Academy Press.',
        'Kaufman, M. B., Brodin, K. A., & Sarafian, A. (2005, April/May). Effect of prescriber education on the use of medications contraindicated in older adults in a managed Medicare population. Journal of Managed Care & Specialty Pharmacy, 11(3), 211-219.',
        'Lau, D.T., J.D., Kasper, D.E., Potter, A. Lyles. (2004). Potentially Inappropriate Medication Prescriptions Among Elderly Nursing Home Residents: Their Scope and Associated Resident and Facility Characteristics. Health Services Research, 39(5), 1257-1276.',
        'MacKinnon, N. J., & Hepler, C. D. (2003). Indicators of preventable drug-related morbidity in older adults: Use within a managed care organization. Journal of Managed Care & Specialty Pharmacy, 9(2), 134-141.',
        'Rothberg, M. B., Perkow, P. S., Liu, F., et al. (2008). Potentially inappropriate medication use in hospitalized elders. Journal of Hospital Medicine, 3(2), 91-102.',
        'Zhan, C., Sangl, J., Bierman, A. S., et al. (2001). Potentially inappropriate medication use in the community-dwelling elderly. JAMA, 286(22), 2823-2868.',
        '2019 American Geriatrics Society Beers Criteria Update Expert Panel. (2019). American Geriatrics Society 2019 Updated AGS Beers Criteria for Potentially Inappropriate Medication Use in Older Adults. Journal of the American Geriatrics Society, 67(4), 674-694.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 65 years and older who had a visit during the measurement
        period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with at least two orders for the same high-risk medication on different
        days during the measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The measure is based on recommendations from the American
        Geriatrics Society Beers Criteria[R] for Potentially Inappropriate Medication Use in Older
        Adults (2019 Update). The criteria were developed through key clinical expert consensus
        processes by Beers in 1997, Zhan in 2001 and an updated process by Fick et al. in 2003,
        2012, 2015, and 2019. The Beers Criteria identifies lists of drugs that are potentially
        inappropriate for all older adults and drugs that are potentially inappropriate in older
        adults based on various high-risk factors such as dosage, days supply and underlying
        diseases or conditions.
        NCQA's Geriatric Measurement Advisory Panel recommended a subset of drugs that should be
        used with caution in older adults for inclusion in the measure based upon the
        recommendations in the Beers Criteria.
        """
        pass



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AceInhibitorOrArbOrArni, AcuteInpatient, AnnualWellnessVisit,
    CareServicesInLongTermResidentialFacility, DementiaMedications, Diabetes, DiabeticNephropathy,
    DialysisEducation, Ed, EncounterInpatient, EsrdMonthlyOutpatientServices, Ethnicity,
    FrailtyDevice, FrailtyDiagnosis, FrailtyEncounter, FrailtySymptom,
    GlomerulonephritisAndNephroticSyndrome, HomeHealthcareServices, HospiceCareAmbulatory,
    HypertensiveChronicKidneyDisease, KidneyFailure, NonacuteInpatient, NursingFacilityVisit,
    Observation, OfficeVisit, OncAdministrativeSex, OtherServicesRelatedToDialysis, Outpatient,
    Payer, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Proteinuria, Race, UrineProteinTests)


class ClinicalQualityMeasure134v9(ClinicalQualityMeasure):
    """
    Diabetes: Medical Attention for Nephropathy

    Description: The percentage of patients 18-75 years of age with diabetes who had a nephropathy
    screening test or evidence of nephropathy during the measurement period

    Definition: None

    Rationale: As the seventh leading cause of death in the U.S., diabetes kills approximately
    79,500 people a year and affects more than 30 million Americans (9.4 percent of the U.S.
    population) (CDC, 2017a, 2017b). Diabetes is a long-lasting disease marked by high blood
    glucose levels, resulting from the body's inability to produce or use insulin properly (CDC,
    2019). People with diabetes are at increased risk of serious health complications including
    vision loss, heart disease, stroke, kidney failure, amputation of toes, feet or legs, and
    premature death. (CDC, 2016).

    In 2017, diabetes cost the U.S. an estimated $327 billion: $237 billion in direct medical costs
    and $90 billion in reduced productivity. This is a 34 percent increase from the estimated $245
    billion spent on diabetes in 2012 (American Diabetes Association, 2018).

    High blood sugar levels in patients with diabetes put them at a higher risk of damaging their
    kidneys and causing chronic kidney disease, which can lead to kidney failure (CDC, 2016,
    2017c). During 2011-2012 there were 36.5% new cases of chronic kidney disease (stages 1-4)
    among 297,000 diabetic patients 20 years and older (Murphy et al., 2016). In 2014, diabetes
    accounted for 44% of 118,000 new cases of end stage renal disease (United States Renal Data
    System, 2016).

    Guidance: Only patients with a diagnosis of Type 1 or Type 2 diabetes should be included in the
    denominator of this measure; patients with a diagnosis of secondary diabetes due to another
    condition should not be included.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms134v9
    """

    title = 'Diabetes: Medical Attention for Nephropathy'

    identifiers = ['CMS134v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Association of Clinical Endocrinologists & American College of Endocrinology. (2015). Clinical practice guidelines for developing a diabetes mellitus comprehensive care plan—2015. Endocrine Practice, 21(Suppl. 1). Retrieved from https://www.aace.com/files/dm-guidelines-ccp.pdf',
        'American Diabetes Association. (2018). Economic costs of diabetes in the U.S. in 2017. Diabetes Care, 41, 917-928. Retrieved from http://care.diabetesjournals.org/content/early/2018/03/20/dci18-0007',
        'American Diabetes Association. (2019). 11. Microvascular complications and foot care: Standards of medical care in diabetes—2019. Diabetes Care, 42(Suppl. 1), S124-S138. https://doi.org/10.2337/dc19-S011',
        'Centers for Disease Control and Prevention. (2016). At a glance 2016: Diabetes—Working to reverse the U.S. epidemic. Atlanta, GA: Author. Retrieved from https://upcap.org/admin/wp-content/uploads/2016/06/Diabetes-at-a-Glance.pdf',
        'Centers for Disease Control and Prevention. (2017a). Health, United States, 2016: With chartbook on long-term trends in health. Retrieved from https://www.cdc.gov/nchs/data/hus/hus16.pdf',
        'Centers for Disease Control and Prevention. (2017b). National diabetes statistics report, 2017. Atlanta, GA: U.S. Department of Health and Human Services, CDC. Retrieved from https://www.cdc.gov/diabetes/pdfs/data/statistics/national-diabetes-statistics-report.pdf',
        'Centers for Disease Control and Prevention. (2019). About diabetes. Retrieved from https://www.cdc.gov/diabetes/basics/diabetes.html',
        'Centers for Disease Control and Prevention. (2017c). National chronic kidney disease fact sheet. Retrieved from https://www.cdc.gov/diabetes/pubs/pdf/kidney_factsheet.pdf',
        'Murphy, D., McCulloch, C. E., Lin, F., et al. (2016). Trends in prevalence of chronic kidney disease in the United States. Annals of Internal Medicine, 165(7), 473-481. Retrieved from https://annals.org/aim/fullarticle/2540849/trends-prevalence-chronic-kidney-disease-united-states',
        'United States Renal Data System. (2016). 2016 USRDS annual data report: Epidemiology of kidney disease in the United States. Bethesda, MD: National Institute of Diabetes and Digestive and Kidney Diseases, National Institutes of Health.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 18-75 years of age with diabetes with a visit during the
        measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period.

        Exclude patients 66 and older who are living long term in an institution for more than 90
        consecutive days during the measurement period.

        Exclude patients 66 and older with advanced illness and frailty because it is unlikely that
        patients will benefit from the services being measured.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with a screening for nephropathy or evidence of nephropathy during the
        measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: American Diabetes Association (2019):
        Screening
        - At least once a year, assess urinary albumin (e.g., spot urinary albumin-to-creatinine
        ratio [UACR]) and estimated glomerular filtration rate (eGFR) in patients with type 1
        diabetes with duration of >=5 years, in all patients with type 2 diabetes, and in all
        patients with comorbid hypertension. (Level of evidence: B)

        Treatment
        - An angiotensin-converting enzyme (ACE) inhibitor or angiotensin receptor blocker (ARB) is
        not recommended for the primary prevention of diabetic kidney disease in patients with
        diabetes who have normal blood pressure, normal UACR (<30 mg/g creatinine), and normal
        estimated glomerular filtration rate. (Level of evidence: B)
        - In nonpregnant patients with diabetes and hypertension, either an ACE inhibitor or ARB is
        recommended for those with modestly elevated UACR (30-299 mg/g creatinine) (Level of
        evidence: B) and is strongly recommended for those with UACR >=300 mg/g creatinine and/or
        eGFR <60 mL/min/1.73.m2. (Level of evidence: A)
        - Periodically monitor serum creatinine and potassium levels for the development of
        increased creatinine or changes in potassium when ACE inhibitors, ARB, or diuretics are
        used. (Level of evidence: B)
        - Continued monitoring of UACR in patients with albuminuria treated with an ACE inhibitor
        or an ARB is reasonable to assess the response to treatment and progression of chronic
        kidney disease. (Level of evidence: E)
        - When eGFR is <60 mL/min/1.73 m2, evaluate and manage potential complications of chronic
        kidney disease. (Level of evidence: E)
        -Patients should be referred for evaluation for renal replacement treatment if they have an
        eGFR <30 mL/min/1.73 m2. (Level of evidence: A)
        -Promptly refer to a physician experienced in the care of kidney disease for uncertainty
        about the etiology of kidney disease, difficult management issues, and rapidly progressing
        kidney disease. (Level of evidence: B)

        American Association of Clinical Endocrinologists & American College of Endocrinology
        (2015):
        - Beginning 5 years after diagnosis in patients with type 1 diabetes (if diagnosed before
        age 30) or at diagnosis in patients with type 2 diabetes and those with type 1 diabetes
        diagnosed after age 30, annual assessment of serum creatinine to determine the estimated
        glomerular filtration rate (eGFR) and urine albumin excretion rate (AER) should be
        performed to identify, stage, and monitor progression of diabetic nephropathy. (Grade C;
        best evidence level 3)
        - Patients with nephropathy should be counseled regarding the need for optimal glycemic
        control, blood pressure control, dyslipidemia control, and smoking cessation. (Grade B;
        best evidence level 2)
        - In addition, they should have routine monitoring of albuminuria, kidney function
        electrolytes, and lipids. (Grade B; best evidence level 2)
        - Associated conditions such as anemia and bone and mineral disorders should be assessed as
        kidney function declines. (Grade D; best evidence level 4)
        - Referral to a nephrologist is recommended well before the need for renal replacement
        therapy. (Grade D; best evidence level 4)
        """
        pass



from unittest import TestCase

from canvas_workflow_kit.value_set.specials import (
    Covid19QuestionnaireSymptomaticSurveillance,
    Covid19QuestionnaireHighRiskOutreach,
    Hcc005v1AnnualWellnessVisit,
    CMS125v6Tomography,
    CMS130v6CtColonography,
    CMS134v6Dialysis,
    LabReportCreatinine,
    DiabetesWithoutComplication,
    DiabetesEyeConditionSuspect,
    DiabetesEyeClassConditionSuspect,
    DiabetesNeurologicConditionSuspect,
    DiabetesRenalConditionSuspect,
    DiabetesCirculatoryClassConditionSuspect,
    DiabetesOtherClassConditionSuspect,
    DysrhythmiaClassConditionSuspect,
)


class TestHcc005v1AnnualWellnessVisit(TestCase):

    def test_set(self):
        expected = {
            'hcpcs': {'99397', 'G0439', 'G0402', '99387', 'G0438'},
        }
        self.assertEqual(expected, Hcc005v1AnnualWellnessVisit.values)


class TestCMS125v6Tomography(TestCase):

    def test_set(self):
        expected = {
            'loinc': {'72142-3'},
        }
        self.assertEqual(expected, CMS125v6Tomography.values)


class TestCMS130v6CtColonography(TestCase):

    def test_set(self):
        expected = {
            'loinc': {'79101-2'},
        }
        self.assertEqual(expected, CMS130v6CtColonography.values)


class TestCMS134v6Dialysis(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {'Z992'},
            'snomedct': {'2080P0210X', '207RN0300X'},
        }
        self.assertEqual(expected, CMS134v6Dialysis.values)


class TestLabReportCreatinine(TestCase):

    def test_set(self):
        expected = {
            'loinc': {'2160-0'},
        }
        self.assertEqual(expected, LabReportCreatinine.values)


class TestDiabetesWithoutComplication(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {'E119'},
        }
        self.assertEqual(expected, DiabetesWithoutComplication.values)


class TestDiabetesEyeConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {'H28', 'H36'},
        }
        self.assertEqual(expected, DiabetesEyeConditionSuspect.values)


class TestDiabetesEyeClassConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {r'H35\d*'},
        }
        self.assertEqual(expected, DiabetesEyeClassConditionSuspect.values)


class TestDiabetesNeurologicConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {'G63', 'G737', 'G53'},
        }
        self.assertEqual(expected, DiabetesNeurologicConditionSuspect.values)


class TestDiabetesRenalConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {'N181', 'N182', 'N183', 'N184', 'N185', 'N186', 'N189'},
        }
        self.assertEqual(expected, DiabetesRenalConditionSuspect.values)


class TestDiabetesCirculatoryClassConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {r'I73\d*', r'I70\d*', r'I71\d*', r'I79\d*'},
        }
        self.assertEqual(expected, DiabetesCirculatoryClassConditionSuspect.values)


class TestDiabetesOtherClassConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {r'M14\d*', r'L97\d*', r'L984\d*'},
        }
        self.assertEqual(expected, DiabetesOtherClassConditionSuspect.values)


class TestDysrhythmiaClassConditionSuspect(TestCase):

    def test_set(self):
        expected = {
            'icd10cm': {r'I42\d*', r'I47\d*', r'I48\d*', r'I49\d*'},
        }
        self.assertEqual(expected, DysrhythmiaClassConditionSuspect.values)


class TestCovid19QuestionnaireSymptomaticSurveillance(TestCase):

    def test_set(self):
        expected = {
            'canvas': {'CANVAS0001'},
        }
        self.assertEqual(expected, Covid19QuestionnaireSymptomaticSurveillance.values)


class TestCovid19QuestionnaireHighRiskOutreach(TestCase):

    def test_set(self):
        expected = {
            'canvas': {'CANVAS0006'},
        }
        self.assertEqual(expected, Covid19QuestionnaireHighRiskOutreach.values)



from .medication import *
from .diagnosis import *
from .encounter_performed import *
from .encounter import *
from .lab_test import *
from .assessment_performed import *
from .physical_exam import *
from .procedure_performed import *
from .procedure import *
from .device import *
from .communication import *
from .immunization import *
from .allergy import *
from .patient_characteristic import *
from .rationale import *
from .other import *
from .assessment import *
from .result import *



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    DiagnosisOfHypertension, DietaryRecommendations, EncounterToScreenForBloodPressure, Ethnicity,
    FollowUpWithin4Weeks, FollowUpWithinOneYear, LifestyleRecommendation, OncAdministrativeSex,
    Payer, PharmacologicTherapyForHypertension, Race, RecommendationToIncreasePhysicalActivity,
    ReferralOrCounselingForAlcoholConsumption, ReferralToPrimaryCareOrAlternateProvider,
    WeightReductionRecommended)


class ClinicalQualityMeasure22v9(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Screening for High Blood Pressure and Follow-Up Documented

    Description: Percentage of patient visits for patients aged 18 years and older seen during the
    measurement period who were screened for high blood pressure AND a recommended follow-up plan
    is documented, as indicated, if blood pressure is pre-hypertensive or hypertensive

    Definition: Blood Pressure (BP) Classification:
    BP is defined by four (4) BP reading classifications: Normal, Pre-Hypertensive, First
    Hypertensive, and Second Hypertensive Readings

    Recommended BP Follow-Up:
    The Joint National Committee on the Prevention, Detection, Evaluation, and Treatment of High
    Blood Pressure (JNC 7) recommends BP screening intervals, lifestyle modifications and
    interventions based on the current BP reading as listed in the "Recommended Blood Pressure
    Follow-Up Interventions" listed below

    Recommended Lifestyle Modifications:
    The JNC 7 report outlines lifestyle modifications which must include one or more of the
    following as indicated:
      *  Weight Reduction
      *  Dietary Approaches to Stop Hypertension (DASH) Eating Plan
      *  Dietary Sodium Restriction
      *  Increased Physical Activity
      *  Moderation in alcohol (ETOH) Consumption

    Second Hypertensive Reading:
    Requires a BP reading of systolic BP >= 140 mmHg OR diastolic BP >= 90 mmHg during the current
    encounter AND a most recent BP reading within the last 12 months systolic BP >= 140 mmHg OR
    diastolic BP >= 90 mmHg

    Second Hypertensive Reading BP Interventions:
    The JNC 7 report outlines BP follow-up interventions for a second hypertensive BP reading and
    must include one or more of the following as indicated:
      *  Anti-Hypertensive Pharmacologic Therapy
      *  Laboratory Tests
      *  Electrocardiogram (ECG)

    Recommended Blood Pressure Follow-Up Interventions:
      *  Normal BP: No follow-up required for systolic BP < 120 mmHg AND diastolic BP < 80 mmHg
      *  Pre-Hypertensive BP: Follow-up with rescreen every year with systolic BP of 120-139 mmHg
    OR
          diastolic BP of 80-89 mmHg AND recommend lifestyle modifications OR referral to
          Alternative/Primary Care Provider
      *  First Hypertensive BP Reading: Patients with one elevated reading of systolic BP >= 140
    mmHg
          OR diastolic BP >= 90 mmHg:
               *  Follow-up with rescreen > 1 day and < 4 weeks AND recommend lifestyle
    modifications
                   OR referral to Alternative/Primary Care Provider
      *  Second Hypertensive BP Reading: Patients with second elevated reading of systolic
          BP >= 140 mmHg OR diastolic BP >= 90 mmHg:
               *  Follow-up with Recommended lifestyle recommendations AND one or more of the
    Second
                   Hypertensive Reading Interventions OR referral to Alternative/Primary Care
    Provider

    Rationale: Hypertension is a prevalent condition that affects approximately 66.9 million people
    in the United States. It is estimated that about 20-40% of the adult population has
    hypertension; the majority of people over age 65 have a hypertension diagnosis (Appleton SL, et
    al., 2012 and Luehr D, et al., 2012). Winter (2013) noted that 1 in 3 American adults have
    hypertension and the lifetime risk of developing hypertension is 90% (Winter KH, et al., 2013).
    The African American population or non-Hispanic Blacks, the elderly, diabetics and those with
    chronic kidney disease are at increased risk of stroke, myocardial infarction and renal
    disease. Non-Hispanic Blacks have the highest prevalence at 38.6% (Winter KH, et al., 2013).
    Hypertension is a major risk factor for ischemic heart disease, left ventricular hypertrophy,
    renal failure, stroke and dementia (Luehr D, et al., 2012).

    Hypertension is the most common reason for adult office visits other than pregnancy. Garrison
    (2013) stated that in 2007, 42 million ambulatory visits were attributed to hypertension
    (Garrison GM and Oberhelman S, 2013). It also has the highest utilization of prescription
    drugs. Numerous resources and treatment options are available, yet only about 40-50% of the
    hypertensive patients have their blood pressure under control (<140/90) (Appleton SL, et al.,
    2012, Luehr D, et al., 2012). In addition to medication non-compliance, poor outcomes are also
    attributed to poor adherence to lifestyle changes such as a low-sodium diet, weight loss,
    increased exercise and limiting alcohol intake. Many adults find it difficult to continue
    medications and lifestyle changes when they are asymptomatic. Symptoms of elevated blood
    pressure usually do not occur until secondary problems arise such as with vascular diseases
    (myocardial infarction, stroke, heart failure and renal insufficiency) (Luehr D, et al., 2012).

    Appropriate follow-up after blood pressure measurement is a pivotal component in preventing the
    progression of hypertension and the development of heart disease. Detection of marginally or
    fully elevated blood pressure by a specialty clinician warrants referral to a provider familiar
    with the management of hypertension and prehypertension. The 2010 ACCF/AHA Guideline for the
    Assessment of Cardiovascular Risk in Asymptomatic Adults continues to support using a global
    risk score such as the Framingham Risk Score, to assess risk of coronary heart disease (CHD) in
    all asymptomatic adults (Greenland P, et al., 2010). Lifestyle modifications have demonstrated
    effectiveness in lowering blood pressure (JNC 7, 2003). The synergistic effect of several
    lifestyle modifications results in greater benefits than a single modification alone. Baseline
    diagnostic/laboratory testing establishes if a co-existing underlying condition is the etiology
    of hypertension and evaluates if end organ damage from hypertension has already occurred.
    Landmark trials such as ALLHAT have repeatedly proven the efficacy of pharmacologic therapy to
    control blood pressure and reduce the complications of hypertension. Follow-up intervals based
    on blood pressure control have been established by the JNC 7 and the USPSTF.

    Guidance: This eCQM is an episode-based measure and should be reported at every visit for
    patients aged 18 years and older. This measure will be calculated based upon the clinical
    actions performed at every visit during the measurement period for each patient. The measure
    requires that blood pressure measurements (i.e., diastolic and systolic) be obtained during
    each visit in order to determine the blood pressure reading used to evaluate if an intervention
    is needed.

    Both the systolic and diastolic blood pressure measurements are required for inclusion. If
    there are multiple blood pressures obtained during a patient visit, only the last, or most
    recent, pressure measurement will be used to evaluate the measure requirements.

    The intent of this measure is to screen patients for high blood pressure and provide
    recommended follow-up as indicated. The documented follow-up plan must be related to the
    current blood pressure reading as indicated, example: "Patient referred to primary care
    provider for BP management."

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms22v9
    """

    title = 'Preventive Care and Screening: Screening for High Blood Pressure and Follow-Up Documented'

    identifiers = ['CMS22v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Appleton, S. L., Neo, C., Hill, C. L., Douglas, K. A., & Adams, R. J. (2013). Untreated hypertension: prevalence and patient factors and beliefs associated with under-treatment in a population sample. Journal of Human Hypertension, 27, 453-462. doi:10.1038/jhh.2012.62ID',
        'Garrison, G. M. & Oberhelman, S. (2013). Screening for hypertension annually compared with current practice. Annals of Family Medicine, 11 (2), 116-121. doi:10.1370/afm.1467',
        'Greenland, P., Alpert, J. S., Beller, G. A., Benjamin, E. J., Budoff, M. J., Fayad, Z. A., Foster, E., Hlatky, M. A., Hodgson, J. M., Kushner, F. G., Lauer, M. S., Shaw, L. J., Smith, S. C., Jr, Taylor, A. J., Weintraub, W. S., Wenger, N. K., Jacobs, A. K., Smith, S. C., Jr, Anderson, J. L., Albert, N., … American Heart Association (2010). 2010 ACCF/AHA guideline for assessment of cardiovascular risk in asymptomatic adults: a report of the American College of Cardiology Foundation/American Heart Association Task Force on Practice Guidelines. Journal of the American College of Cardiology, 56(25), e50–e103. https://doi.org/10.1016/j.jacc.2010.09.001',
        'Luehr, D., Woolley, T., Burke, R., Dohmen, F., Hayes, R., Johnson, M...., Schoenleber, M. (2012). Hypertension diagnosis and treatment; Institute for Clinical Systems Improvement health care guideline. Updated November, 2012.',
        'U.S. Preventive Services Task Force (USPSTF) (2007). Screening for high blood pressure: U.S. Preventive Services Task Force reaffirmation recommendation statement. Annals of Internal Medicine; 147(11):783-6',
        'U.S. Department of Health and Human Services, National Institutes of Health, National Heart, Lung, and Blood Institute & National High Blood Pressure Education Program (2003). The Seventh Report of the Joint National Committee on the Prevention, Detection, Evaluation, and Treatment of High Blood Pressure (JNC-7). NIH Publication No. 03-5233',
        'Winter, K. H., Tuttle, L. A. & Viera, A.J. (2013). Hypertension. Primary Care Clinics in Office Practice, 40, 179-194. doi:10.1016/j.pop.2012.11.008',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patient visits for patients aged 18 years and older at the
        beginning of the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patient has an active diagnosis of hypertension

        Exceptions: Documentation of medical reason(s) for not screening for high blood pressure
        (e.g., patient is in an urgent or emergent medical situation where time is of the essence
        and to delay treatment would jeopardize the patient's health status).

        Documentation of patient reason(s) for not screening for blood pressure measurements or for
        not ordering an appropriate follow-up intervention if patient is pre-hypertensive or
        hypertensive (e.g., patient refuses).
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patient visits where patients were screened for high blood pressure AND have a
        recommended follow-up plan documented, as indicated, if the blood pressure is pre-
        hypertensive or hypertensive

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The U.S. Preventive Services Task Force (USPSTF) recommends
        screening for high blood pressure in adults age 18 years and older. This is a grade A
        recommendation.
        """
        pass



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    CareServicesInLongTermResidentialFacility, DiabeticRetinopathy, Ethnicity,
    LevelOfSeverityOfRetinopathyFindings, MacularEdemaFindingsPresent, MacularExam,
    NursingFacilityVisit, OfficeVisit, OncAdministrativeSex, OphthalmologicalServices,
    OutpatientConsultation, Payer, Race)


class ClinicalQualityMeasure142v9(ClinicalQualityMeasure):
    """
    Diabetic Retinopathy: Communication with the Physician Managing Ongoing Diabetes Care

    Description: Percentage of patients aged 18 years and older with a diagnosis of diabetic
    retinopathy who had a dilated macular or fundus exam performed with documented communication to
    the physician who manages the ongoing care of the patient with diabetes mellitus regarding the
    findings of the macular or fundus exam at least once within 12 months

    Definition: Communication - May include documentation in the medical record indicating that the
    findings of the dilated macular or fundus exam were communicated (e.g., verbally, by letter)
    with the clinician managing the patient's diabetic care OR a copy of a letter in the medical
    record to the clinician managing the patient's diabetic care outlining the findings of the
    dilated macular or fundus exam.

    Findings - Includes level of severity of retinopathy (e.g., mild nonproliferative, moderate
    nonproliferative, severe nonproliferative, very severe nonproliferative, proliferative) AND the
    presence or absence of macular edema.

    Rationale: Diabetic retinopathy is a prevalent complication of diabetes, estimated to affect
    28.5% of diabetic patients in the US (Zhang et al., 2010). Diabetic Retinopathy is a key
    indicator of systemic complications of diabetes (Zhang, 2010). Coordination of care between the
    eye care specialist and the physician managing a patient’s ongoing diabetes care is essential
    in stemming the progression of vision loss. Communication from the eye care specialist to a
    primary care physician facilitates the exchange of information about the severity and
    progression of a patient’s diabetic retinopathy, adherence to recommended ocular care, need for
    follow-up visits, and treatment plans (Storey, Murchison, Pizzi, Hark, Dai, Leiby & Haller,
    2016). Data from the Diabetes Control and Complications Trial showed that diabetic treatment
    and maintenance of glucose control delays the onset and slows the progression of diabetic
    retinopathy (Aiello & DCCT/EDIC Research Group, 2014).

    Guidance: The measure, as written, does not specifically require documentation of laterality.
    Coding limitations in particular clinical terminologies do not currently allow for that level
    of specificity (ICD-10-CM includes laterality, but SNOMED-CT does not uniformly include this
    distinction). Therefore, at this time, it is not a requirement of this measure to indicate
    laterality of the diagnoses, findings or procedures. Available coding to capture the data
    elements specified in this measure has been provided. It is assumed that the eligible
    professional or eligible clinician will record laterality in the patient medical record, as
    quality care and clinical documentation should include laterality.

    The communication of results to the primary care physician providing ongoing care of a
    patient's diabetes should be completed soon after the dilated exam is performed. Eligible
    professionals or eligible clinicians reporting on this measure should note that all data for
    the reporting year is to be submitted by the deadline established by CMS. Therefore, eligible
    professionals or eligible clinicians who see patients towards the end of the reporting period
    (i.e., December in particular), should communicate the results of the dilated macular exam as
    soon as possible in order for those patients to be counted in the measure numerator.
    Communicating the results as soon as possible after the date of the exam will ensure the data
    are included in the submission to CMS.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms142v9
    """

    title = 'Diabetic Retinopathy: Communication with the Physician Managing Ongoing Diabetes Care'

    identifiers = ['CMS142v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Aiello, L. P., & DCCT/EDIC Research Group (2014). Diabetic retinopathy and other ocular findings in the diabetes control and complications trial/epidemiology of diabetes interventions and complications study. Diabetes care, 37(1), 17–23. doi:10.2337/dc13-2251',
        'American Academy of Ophthalmology. (2017). Diabetic retinopathy Preferred Practice Pattern. San Francisco, CA: American Academy of Ophthalmology.',
        'Storey, P. P., Murchison, A. P., Pizzi, L. T., Hark, L. A., Dai, Y., Leiby, B. E., & Haller, J. A. Impact of physician communication on diabetic eye examination adherence: Results from a Retrospective Cohort Analysis. Retina. 2016 Jan;36(1):20-7. doi:10.1097/IAE.0000000000000652',
        'Zhang, X., Saaddine, J. B., Chou, C. F., Cotch, M. F., Cheng, Y. J., Geiss, L. S., … Klein, R. (2010). Prevalence of diabetic retinopathy in the United States, 2005-2008. JAMA, 304(6), 649–656. doi:10.1001/jama.2010.1111',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 years and older with a diagnosis of diabetic
        retinopathy
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population who had a dilated macular or fundus exam performed

        Exclusions: None

        Exceptions: Documentation of medical reason(s) for not communicating the findings of the
        dilated macular or fundus exam to the physician who manages the ongoing care of the patient
        with diabetes.

        Documentation of patient reason(s) for not communicating the findings of the dilated
        macular or fundus exam to the physician who manages the ongoing care of the patient with
        diabetes.
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients with documentation, at least once within 12 months, of the findings of
        the dilated macular or fundus exam via communication to the physician who manages the
        patient's diabetic care

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The ophthalmologist should refer patients with diabetes to a
        primary care physician for appropriate management of their systemic condition and should
        communicate examination results to the physician managing the patient's ongoing diabetes
        care (III; Good Quality; Strong Recommendation) (American Academy of Ophthalmology, 2017).

        Ophthalmologists should communicate the ophthalmologic findings and level of retinopathy
        with the primary care physician as well as the need for optimizing metabolic control (III;
        Good Quality; Strong Recommendation) (American Academy of Ophthalmology, 2017).

        Close partnership with the primary care physician is important to make sure that the care
        of the patient is optimized (III; Good Quality; Strong Recommendation) (American Academy of
        Ophthalmology, 2017).
        """
        pass



# type: ignore
from canvas_workflow_kit.builtin_cqms.cms138v6_preventive_care_and_screening_tobacco_use_screening_and_cessation_intervention import (  # noqa: E501
    STATUS_DUE,
    STATUS_NOT_APPLICABLE,
    ClinicalQualityMeasure138v6,
    InterviewRecommendation,
    ProtocolResult,
    TobaccoUseScreening,
    events
)


class ClinicalQualityMeasure138v6p1(ClinicalQualityMeasure138v6):
    """
    Use only the Population 1's path of the ClinicalQualityMeasure138v6
    """

    class Meta:
        title = 'Preventive Care and Screening: Tobacco Use: Screening'
        version = '2019-04-18v1'
        description = 'Patients aged 18 years and older who have not been screened for tobacco use in the last year.'  # noqa: E501
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS138v6.html'

        identifiers = ['CMS138v6p1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure138v6.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure138v6.CHANGE_BILLING_LINE_ITEM,
            ClinicalQualityMeasure138v6.CHANGE_INTERVIEW,
            ClinicalQualityMeasure138v6.CHANGE_PATIENT,
        ]

    def in_initial_population(self) -> bool:
        super().in_initial_population()
        return self._populations[self.POPULATION_1].in_initial_population

    def in_denominator(self) -> bool:
        super().in_denominator()
        return self._populations[self.POPULATION_1].in_denominator

    def in_numerator(self) -> bool:
        super().in_numerator()
        return self._populations[self.POPULATION_1].in_numerator

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            self.in_numerator()
            if (self._populations[self.POPULATION_1].in_denominator and
                    not self._populations[self.POPULATION_1].in_numerator):
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(
                    f'{self.patient.first_name} should be screened for tobacco use.')
                result.add_recommendation(
                    InterviewRecommendation(
                        key='CMS138v6p1_RECOMMEND_TOBACCO_USE_SCREENING',
                        rank=1,
                        button='Plan',
                        patient=self.patient,
                        questionnaires=[TobaccoUseScreening],
                        title='Complete tobacco use questionnaire'))

            elif self.tobacco_use_screening_non_user:
                self.satisfied_result(
                    self.tobacco_use_screening_non_user,
                    '{name} had a Tobacco screening {date} and is not a smoker.', result)

            elif self.tobacco_use_screening_user:
                self.satisfied_result(self.tobacco_use_screening_user,
                                      '{name} had a Tobacco screening {date} and is a smoker.',
                                      result)
        elif self.patient.age_at(self.timeframe.end) < self.MINIMUM_AGE:
            result.status = STATUS_NOT_APPLICABLE
            result.due_in = (
                self.patient.birthday.shift(years=self.MINIMUM_AGE) - self.timeframe.end).days
        return result



from typing import Dict, List

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.specials import KidneyFailure, LabReportCreatinine
from canvas_workflow_kit.value_set.v2018 import HypertensiveChronicKidneyDisease


class Hcc002v2(ClinicalQualityMeasure):

    class Meta:
        title = 'CKD suspect'

        version = '2019-02-12v1'

        description = ('All patients with evidence of two or more elevated eGFR values '
                       'and no active CKD problem on the Conditions List.')
        information = ('https://canvas-medical.zendesk.com/hc/en-us/articles/'
                       '360059083713-CKD-Suspect-HCC002v2')  # noqa: E501

        identifiers = ['HCC002v2']

        types = ['HCC']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        references = [(
            'Canvas Medical HCC, '
            'https://canvas-medical.zendesk.com/hc/en-us/articles/360059083713-CKD-Suspect-HCC002v2'  # noqa: E501
        )]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_LAB_REPORT,
            ClinicalQualityMeasure.CHANGE_PATIENT,
        ]

    @property
    def high_creatine_levels(self) -> List[Dict]:
        patient_age = int(self.patient.age_at(self.now))

        # we can't calculate an eGFR if age = 0 due to a ZeroDivisionError; we'd need to use a
        # pedatric eGFR instead (and probably should)
        if patient_age == 0:
            return []

        if self.period_adjustment:
            period = self.timeframe
        else:
            period = Timeframe(start=self.timeframe.end.shift(years=-2), end=self.timeframe.end)

        creatinine_lab_reports = (self.patient
                                  .lab_reports
                                  .find(LabReportCreatinine)
                                  .within(period))  # yapf: disable

        return [
            {
                'date': r['originalDate'],
                'value': self.eGFR(self.relative_float(r['value']), r['units']),
            }
            for r in creatinine_lab_reports
            if (self.relative_float(r['value']) > 0 and
                self.eGFR(self.relative_float(r['value']), r['units']) < 60)
        ]  # yapf: disable

    @property
    def has_active_condition(self) -> bool:
        return any(True for item in (self.patient
                                     .conditions
                                     .find(HypertensiveChronicKidneyDisease | KidneyFailure))
                   if item['clinicalStatus'] == 'active')  # yapf: disable

    def eGFR(self, creatinine: float, units: str) -> float:
        sex = 0.742 if self.patient.is_female else 1
        race = 1.210 if self.patient.is_african_american else 1
        coefficient = 1 if units == 'mg/dL' else 88.4
        return 186 * pow(creatinine / coefficient, -1.154) * pow(
            int(self.patient.age_at(self.now)), -0.203) * sex * race

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        """
        Patients with 2*eGFR lab value of < 60 in the last 2 years
        """
        return len(self.high_creatine_levels) >= 2

    def in_numerator(self) -> bool:
        """
        Patients with no active Condition on problem list for range of ICD10 Codes
        """
        return self.has_active_condition

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.status = STATUS_SATISFIED
            else:
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(
                    f'{self.patient.first_name} has at least two eGFR measurements < 60 ml/min '
                    'over the last two years suggesting renal disease. '
                    'There is no associated condition on the Conditions List.')
                title = ('Consider updating the Conditions List to include kidney '
                         'related problems as clinically appropriate')
                result.add_recommendation(
                    Recommendation(
                        key='HCC002v2_RECOMMEND_DIAGNOSE',
                        rank=1,
                        button='Diagnose',
                        title=title,
                        narrative=result.narrative,
                        command={'key': 'diagnose'}))
        return result



from unittest import TestCase
from canvas_workflow_kit.value_set.value_set import *


class SuperValueSetTest(TestCase):
    class ValSetA(ValueSet):
        OID = '1.2.3.4'
        VALUE_SET_NAME = 'Set A'
        CPT = {'A_abc', 'A_123'}
        SNOMEDCT = {'A_12345', 'A_23456', 'A_34567'}

    class ValSetB(ValueSet):
        OID = '2.3.4'
        VALUE_SET_NAME = 'Set B'
        CPT = {'B_abc', 'B_123'}
        ICD10CM = {'B_12345', 'B_23456', 'B_34567'}

    class ValSetC(ValueSet):
        OID = '2.5.6'
        VALUE_SET_NAME = 'Set B'
        CPT = {'C_abc', 'C_123'}
        ICD10CM = {'C_12345', 'C_23456', 'C_34567'}

    def test_def(self):
        value_sets = [self.ValSetA, self.ValSetB]
        result = SuperValueSet(value_sets)

        self.assertEqual(len(value_sets), len(result.value_sets))
        if not (result.value_sets[0] is self.ValSetA):
            self.assertIs(self.ValSetB, result.value_sets[0])
        if not (result.value_sets[1] is self.ValSetA):
            self.assertIs(self.ValSetB, result.value_sets[1])

    def test_values(self):
        value_sets = [self.ValSetA, self.ValSetB]
        result = SuperValueSet(value_sets)

        self.assertIsInstance(result.values, defaultdict)
        result = result.values

        expected = {
            'cpt': {'B_123', 'A_123', 'A_abc', 'B_abc'},
            'snomedct': {'A_12345', 'A_34567', 'A_23456'},
            'icd10cm': {'B_12345', 'B_34567', 'B_23456'}
        }
        self.assertEqual(len(expected), len(result))
        for key, val in expected.items():
            self.assertTrue(key in result)
            self.assertEqual(sorted(list(val)), sorted(list(result[key])))

    def test_name(self):
        value_sets = [self.ValSetA, self.ValSetB]
        result = SuperValueSet(value_sets)
        expected = 'Set A or Set B'
        self.assertEqual(expected, result.name)

    def test___or__(self):
        value_sets = [self.ValSetA, self.ValSetB]
        sup_ab = SuperValueSet(value_sets)
        value_sets = [self.ValSetC, self.ValSetB]
        sup_cb = SuperValueSet(value_sets)

        result = sup_ab | sup_cb
        self.assertIsInstance(result, SuperValueSet)

        expected = {
            'cpt': {'C_123', 'B_abc', 'C_abc', 'B_123', 'A_123', 'A_abc'},
            'snomedct': {'A_23456', 'A_34567', 'A_12345'},
            'icd10cm': {'B_34567', 'C_23456', 'C_12345', 'B_23456', 'C_34567', 'B_12345'}
        }
        result = result.values
        self.assertEqual(len(expected), len(result))
        for key, val in expected.items():
            self.assertTrue(key in result)
            self.assertEqual(sorted(list(val)), sorted(list(result[key])))


class ValueSystemsTest(TestCase):

    def test_values(self):
        # all systems + one unknown
        class FullValSet(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'A_abc', 'A_123'}
            CVX = {'B_12345', 'B_23456', 'B_34567'}
            HCPCS = {'C_abc', 'C_123'}
            ICD10CM = {'D_12345', 'D_23456', 'D_34567'}
            ICD10PCS = {'E_abc', 'E_123'}
            ICD9CM = {'F_12345', 'F_23456', 'F_34567'}
            LOINC = {'G_abc', 'G_123'}
            RXNORM = {'H_12345', 'H_23456', 'H_34567'}
            SNOMEDCT = {'I_12345', 'I_23456', 'I_34567'}
            ZZZZZ = {'K_12345', 'K_23456', 'K_34567'}

        expected = {
            'cpt': {'A_123', 'A_abc'},
            'cvx': {'B_34567', 'B_23456', 'B_12345'},
            'hcpcs': {'C_123', 'C_abc'},
            'icd10cm': {'D_34567', 'D_12345', 'D_23456'},
            'icd10pcs': {'E_123', 'E_abc'},
            'icd9cm': {'F_12345', 'F_34567', 'F_23456'},
            'loinc': {'G_abc', 'G_123'},
            'rxnorm': {'H_12345', 'H_23456', 'H_34567'},
            'snomedct': {'I_34567', 'I_12345', 'I_23456'}
        }
        result = FullValSet.values
        self.assertEqual(len(expected), len(result))
        for key, val in expected.items():
            tmp = False
            if key in result:
                self.assertEqual(sorted(list(val)), sorted(list(result[key])))
                tmp = True
            self.assertTrue(tmp)

        # only 2 systems
        class PartValSet(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'A_abc', 'A_123'}
            SNOMEDCT = {'I_12345', 'I_23456', 'I_34567'}
            ZZZZZ = {'K_12345', 'K_23456', 'K_34567'}

        expected = {
            'cpt': {'A_123', 'A_abc'},
            'snomedct': {'I_34567', 'I_12345', 'I_23456'}
        }
        result = PartValSet.values
        self.assertEqual(len(expected), len(result))
        for key, val in expected.items():
            tmp = False
            if key in result:
                self.assertEqual(sorted(list(val)), sorted(list(result[key])))
                tmp = True
            self.assertTrue(tmp)

        # no system
        class PartValSet(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            ZZZZZ = {'K_12345', 'K_23456', 'K_34567'}

        expected = {}
        result = PartValSet.values
        self.assertEqual(len(expected), len(result))

    def test_name(self):
        class PartValSet(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            ZZZZZ = {'K_12345', 'K_23456', 'K_34567'}

        expected = 'Reco for the unit tests'
        result = PartValSet.name
        self.assertEqual(expected, result)

    def test___or__(self):
        class ValSetA(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Set A'
            CPT = {'A_abc', 'A_123'}
            SNOMEDCT = {'A_12345', 'A_23456', 'A_34567'}

        class ValSetB(ValueSet):
            OID = '2.3.4'
            VALUE_SET_NAME = 'Set B'
            CPT = {'B_abc', 'B_123'}
            ICD10CM = {'B_12345', 'B_23456', 'B_34567'}

        expected = {
            'cpt': {'B_123', 'A_123', 'B_abc', 'A_abc'},
            'snomedct': {'A_34567', 'A_23456', 'A_12345'},
            'icd10cm': {'B_34567', 'B_23456', 'B_12345'}
        }
        result = ValSetA | ValSetB
        self.assertIsInstance(result, SuperValueSet)

        result = result.values
        self.assertEqual(len(expected), len(result))
        for key, val in expected.items():
            self.assertTrue(key in result)
            self.assertEqual(sorted(list(val)), sorted(list(result[key])))


class ValueSetTest(TestCase):

    def test_value_system(self):
        class UnitTestReco(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {'12345', '23456', '34567'}

        rec = UnitTestReco()
        expected = ['CANVAS', 'CPT', 'CVX', 'HCPCS', 'ICD10CM', 'ICD10PCS', 'ICD9CM', 'LOINC', 'RXNORM', 'SNOMEDCT', 'FDB', 'INTERNAL', ]
        result = rec.value_systems
        self.assertEqual(expected, result)



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    CarrierOfPredominantlySexuallyTransmittedInfection, Chlamydia, ChlamydiaScreening,
    ComplicationsOfPregnancyChildbirthAndThePuerperium, ContraceptiveMedications,
    DiagnosticStudiesDuringPregnancy, EncounterInpatient, Ethnicity, Female, GenitalHerpes,
    GonococcalInfectionsAndVenerealDiseases, Hiv, HomeHealthcareServices, HospiceCareAmbulatory,
    InflammatoryDiseasesOfFemaleReproductiveOrgans, Isotretinoin, OfficeVisit,
    OncAdministrativeSex, OtherFemaleReproductiveConditions, Payer,
    PreventiveCareEstablishedOfficeVisit0To17, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit0To17, PreventiveCareServicesInitialOfficeVisit18AndUp,
    Race, SexuallyActive, Syphilis, XRayStudyAllInclusive)


class ClinicalQualityMeasure153v9(ClinicalQualityMeasure):
    """
    Chlamydia Screening for Women

    Description: Percentage of women 16-24 years of age who were identified as sexually active and
    who had at least one test for chlamydia during the measurement period

    Definition: None

    Rationale: Chlamydia trachomatis is the most common sexually transmitted bacterial infection in
    the U.S., resulting in roughly 1.7 million cases (Centers for Disease Control and Prevention,
    2018). Chlamydia infections are often asymptomatic, but, if left untreated, can lead to serious
    and irreversible complications (U.S. Preventive Services Task Force, 2014; Centers for Disease
    Control and Prevention, 2018).

    Women are particularly vulnerable when infected with chlamydia. Left untreated, chlamydia can
    cause pelvic inflammatory disease (PID), which can lead to chronic pelvic pain or infertility.
    Pregnant women may also transmit the infection to their infant, potentially resulting in
    neonatal pneumonia (Centers for Disease Control and Prevention, 2018).

    Guidance: Codes to identify sexually active women include codes for: pregnancy, sexually
    transmitted infections, contraceptives or contraceptive devices, and infertility treatments.

    The denominator exclusion does not apply to patients who qualify for the initial population
    (IP) based on services other than the pregnancy test alone. These other services include
    services for sexually transmitted infections, contraceptives or contraceptive devices and
    infertility treatments. For example, a patient who has both a pregnancy test and a chlamydia
    diagnosis, either of which would qualify them for the IP, would not be eligible for this
    denominator exclusion.

    Patient self-report for procedures as well as diagnostic studies should be recorded in
    'Procedure, Performed' template or 'Diagnostic Study, Performed' template in QRDA-1. Patient
    self-report is not allowed for laboratory tests.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms153v9
    """

    title = 'Chlamydia Screening for Women'

    identifiers = ['CMS153v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Centers for Disease Control and Prevention (CDC). (2019). Sexually Transmitted Disease Surveillance 2018. Atlanta: U.S. Department of Health and Human Services.',
        'LeFevre, M. L., & U.S. Preventive Services Task Force. (2014). Screening for Chlamydia and gonorrhea: U.S. Preventive Services Task Force recommendation statement. Annals of Internal Medicine, 161(12), 902–910. https://doi.org/10.7326/M14-1981',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Women 16 to 24 years of age who are sexually active and who had a visit
        in the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Women who are only eligible for the initial population due to a pregnancy test
        and who had an x-ray or an order for a specified medication within 7 days of the pregnancy
        test.

        Exclude patients whose hospice care overlaps the measurement period.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Women with at least one chlamydia test during the measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: U.S. Preventive Services Task Force (2014):
        The task force recommends screening for chlamydia in sexually active females aged 24 years
        or younger and in older women who are at increased risk for infection (B recommendation)
        """
        pass



from ..value_set import ValueSet


class DtapVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent diphtheria, tetanus, and whooping cough (pertussis) (DTaP) vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with diphtheria, tetanus, and whooping cough (pertussis) (DTaP) vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1214'
    VALUE_SET_NAME = 'DTaP Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '106',
        '107',
        '110',
        '120',
        '20',
        '50'
    }

class HepatitisAVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent hepatitis A vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all relevant hepatitis A vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1215'
    VALUE_SET_NAME = 'Hepatitis A Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '31',
        '83',
        '85'
    }

class HepatitisBVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent hepatitis B vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all relevant hepatitis B vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1216'
    VALUE_SET_NAME = 'Hepatitis B Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '08',
        '110',
        '44',
        '45',
        '51'
    }

class HibVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent haemophilus influenzae type b (Hib) vaccines (3-dose schedule).

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with haemophilus influenzae type b (Hib) vaccine codes (3-dose schedule). This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.110.12.1083'
    VALUE_SET_NAME = 'Hib Vaccine (3 dose schedule)'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '49'
    }

class HibVaccine4DoseSchedule(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent haemophilus influenzae type b (Hib) vaccines (4-dose schedule).

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with haemophilus influenzae type b (Hib) vaccine codes (4-dose schedule). This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.110.12.1085'
    VALUE_SET_NAME = 'Hib Vaccine (4 dose schedule)'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '120',
        '148',
        '48',
        '50',
        '51'
    }

class InactivatedPolioVaccineIpv(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent inactivated polio vaccines (IPV).

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all inactivated polio vaccine (IPV) codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1219'
    VALUE_SET_NAME = 'Inactivated Polio Vaccine (IPV)'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '10',
        '110',
        '120',
        '89'
    }

class InfluenzaVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent influenza vaccine codes.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with influenza vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1218'
    VALUE_SET_NAME = 'Influenza Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '135',
        '140',
        '141',
        '150',
        '153',
        '155',
        '158',
        '161',
        '88'
    }

class InfluenzaVaccine_1254(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent influenza vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Immunization.

    **Inclusion Criteria:** Includes only relevant concepts associated with CVX concepts related to influenza vaccines.

    **Exclusion Criteria:** Excludes HCPCS, SNOMED CT and CPT influenza vaccine codes.
    """

    OID = '2.16.840.1.113883.3.526.3.1254'
    VALUE_SET_NAME = 'Influenza Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '135',
        '140',
        '141',
        '144',
        '150',
        '155',
        '158',
        '161',
        '166',
        '168',
        '171',
        '185',
        '186',
        '197',
        '88'
    }

class MeaslesMumpsAndRubellaMmrVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent measles, mumps and rubella (MMR) vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all relevant measles, mumps and rubella (MMR) Vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1224'
    VALUE_SET_NAME = 'Measles, Mumps and Rubella (MMR) Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '03',
        '94'
    }

class PneumococcalConjugateVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent pneumococcal vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all pneumococcal vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1221'
    VALUE_SET_NAME = 'Pneumococcal Conjugate Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '133',
        '152'
    }

class PneumococcalVaccine(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent conjugate and polysaccharide pneumococcal vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with semantic clinical drugs (concepts associated with component, form and strength); generic; prescribable. Includes only relevant concepts associated with the pneumococcal conjugate 13-valent and the pneumococcal polysaccharide 23-valent vaccine. This is a grouping of CVX codes.

    **Exclusion Criteria:** Excludes codes that represent the pneumococcal conjugate 7-valent vaccine.
    """

    OID = '2.16.840.1.113883.3.464.1003.110.12.1027'
    VALUE_SET_NAME = 'Pneumococcal Vaccine'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '109',
        '133',
        '152',
        '33'
    }

class RotavirusVaccine2DoseScheduleAdministered(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent rotavirus (2-dose schedule) vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all rotavirus (2-dose schedule) vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.110.12.1048'
    VALUE_SET_NAME = 'Rotavirus Vaccine (2 dose schedule) Administered'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CPT = {
        '90681'
    }

    SNOMEDCT = {
        '434741000124104'
    }

class RotavirusVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent rotavirus (3-dose schedule) vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all rotavirus (3-dose schedule) vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1223'
    VALUE_SET_NAME = 'Rotavirus Vaccine (3 dose schedule)'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '116',
        '122'
    }

class RotavirusVaccine3DoseScheduleAdministered(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent rotavirus (3-dose schedule) vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all rotavirus (3-dose schedule) vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.110.12.1047'
    VALUE_SET_NAME = 'Rotavirus Vaccine (3 dose schedule) Administered'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CPT = {
        '90680'
    }

    SNOMEDCT = {
        '434731000124109'
    }

class VaricellaZosterVaccineVzv(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent varicella zoster vaccines.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Immunization, Administered.

    **Inclusion Criteria:** Includes only relevant concepts associated with all relevant varicella zoster vaccine codes. This is a grouping of CVX codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.196.12.1170'
    VALUE_SET_NAME = 'Varicella Zoster Vaccine (VZV)'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    CVX = {
        '21',
        '94'
    }


from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.recommendation import InterviewRecommendation
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    AnnualWellnessVisit, AudiologyVisit, CareServicesInLongTermResidentialFacility,
    DischargedToHealthCareFacilityForHospiceCare, DischargedToHomeForHospiceCare,
    EncounterInpatient, Ethnicity, FaceToFaceInteraction, FallsScreening, HomeHealthcareServices,
    HospiceCareAmbulatory, NursingFacilityVisit, OfficeVisit, OncAdministrativeSex,
    OphthalmologicalServices, PatientNotAmbulatory, Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesIndividualCounseling, PreventiveCareServicesInitialOfficeVisit18AndUp,
    Race)


class ClinicalQualityMeasure139v6(ClinicalQualityMeasure):
    """
    Falls: Screening for Future Fall Risk

    Description: Percentage of patients 65 years of age and older who were screened for future fall
    risk during the measurement period

    Definition: Screening for Future Fall Risk: Assessment of whether an individual has experienced
    a fall or problems with gait or balance.  A specific screening tool is not required for this
    measure, however potential screening tools include the Morse Fall Scale and the timed Get-Up-
    And-Go test.

    Fall: A sudden, unintentional change in position causing an individual to land at a lower
    level, on an object, the floor, or the ground, other than as a consequence of sudden onset of
    paralysis, epileptic seizure, or overwhelming external force.

    Rationale: As the leading cause of both fatal and nonfatal injuries for older adults, falls are
    one of the most common and significant health issues facing people aged 65 years or older
    (Schneider, Shubert and Harmon 2010). Moreover, the rate of falls increases with age (Dykes et
    al. 2010). Older adults are five times more likely to be hospitalized for fall-related injuries
    than any other cause-related injury. It is estimated that one in every three adults over 65
    will fall each year (Centers for Disease Control and Prevention 2015). In those over age 80,
    the rate of falls increases to fifty percent (Doherty et al. 2009). Falls are also associated
    with substantial cost and resource use, approaching $30,000 per fall hospitalization (Woolcott
    et al. 2011). Identifying at-risk patients is the most important part of management, as
    applying preventive measures in this vulnerable population can have a profound effect on public
    health (al-Aama 2011). Family physicians have a pivotal role in screening older patients for
    risk of falls, and applying preventive strategies for patients at risk (al-Aama 2011).

    Guidance: None

    More information: https://ecqi.healthit.gov/ecqm/measures/cms139v6
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'National Committee for Quality Assurance',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'al-Aama, T. 2011. "Falls in the Elderly: Spectrum and Prevention." Can Fam Physician 57(7):771-6.',
        'Centers for Disease Control and Prevention. 2015. "Important Facts about Falls" (December 14, 2015) http://www.cdc.gov/HomeandRecreationalSafety/Falls/adultfalls.html',
        'Doherty, M., and J. Crossen-Sills. 2009. "Fall Risk: Keep your patients in balance." The Nurse Practitioner: The American Journal of Primary Health Care 34(12):46-51.',
        'American Geriatrics Society, British Geriatrics Society, and American Academy of Orthopaedic Surgeons Panel on Falls Prevention: Guideline for the prevention of falls in older persons. Journal of the American Geriatrics Society. 2001; 49: 664-672.',
        'National Institute for Clinical Excellence (NICE). Falls: the assessment and prevention of falls in older people. November 2004; clinical guideline 21. Available at: https://www.nice.org.uk/guidance/cg161',
        'Dykes, P.C., D.L. Carroll DL, A. Hurley A, S. Lipsitz S, A. Benoit A, F. Chang F, S. Meltzer S, R. Tsurikova R, L. Zuyov L, B. Middleton B. 2010. "Fall Prevention in Acute Care Hospitals: A Randomized Trial." JAMA . 2010;304(17):1912-1918.',
        'Schneider, E.C., T.E. Shubert, and K.J. Harmon. 2010. "Addressing the Escalating Public Health Issue of Falls Among Older Adults." NC Med J 71(6):547-52.',
        'Woolcott, J.C., K.M. Khan, S. Mitrovic, A.H. Anis, C.A. Marra. 2011. "The Cost of Fall Related Presentations to the ED: A Prospective, In-Person, Patient-Tracking Analysis of Health Resource Utilization." Osteporos Int [Epub ahead of print].',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients aged 65 years and older with a visit during the measurement
        period
        """
        age = self.patient.age_at(self.timeframe.end)
        if age < 65:
            return False

        # FIX: how do we check for a visit within the period?

        return True

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year.

        Exclude patients who were assessed to be non-ambulatory during the measurement period.

        Exceptions: None
        """
        # FIX: is this correct, are these hospice value-sets conditions?
        patient = self.patient
        timeframe = self.timeframe
        if (patient.conditions.before(
                timeframe.end).find(HospiceCareAmbulatory | DischargedToHomeForHospiceCare |
                                    DischargedToHealthCareFacilityForHospiceCare |
                                    PatientNotAmbulatory)):
            return False

        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients who were screened for future fall risk at least once within the
        measurement period

        Exclusions: Not Applicable
        """

        if not self.patient.interviews.find(FallsScreening).within_one_year(self.timeframe):
            return False

    def compute_results(self):
        """
        Clinical recommendation: All older persons who are under the care of a heath professional
        (or their caregivers) should be asked at least once a year about falls.  (AGS/BGS/AAOS)

        Older persons who present for medical attention because of a fall, report recurrent falls
        in the past year, or demonstrate abnormalities of gait and/or balance should have a fall
        evaluation performed.  This evaluation should be performed by a clinician with appropriate
        skills and experience, which may necessitate referral to a specialist (eg, geriatrician).
        (AGS/BGS/AAOS)

        Older people in contact with health care professionals should be asked routinely whether
        they have fallen in the past year and asked about the frequency, context, and
        characteristics of the falls.  (NICE) (Grade C)

        Older people reporting a fall or considered at risk of falling should be observed for
        balance and gait deficits and considered for their ability to benefit from interventions to
        improve strength and balance.  (NICE) (Grade C)
        """
        patient = self.patient

        recommendations = []
        status = 'not_applicable'
        narrative = ''

        if not self.patient.interviews.find(FallsScreening).within_one_year(self.timeframe):
            # should get a screening
            rec = InterviewRecommendation(patient=patient, questionnaire=FallsScreening)
            recommendations.append(rec)

            status = 'due'
            narrative = rec.narrative
        else:
            # they got a screening, all is well
            status = 'current'
            narrative = f"{patient.first_name} was screened for fall risk."

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}



from .internal.attrdict import AttrDict


def assign_meta(new_class, bases, meta):
    m = {}
    for base in bases:
        m.update({k: v for k, v in getattr(base, "_meta", {}).items()})

    m.update({k: v for k, v in getattr(meta, "__dict__", {}).items() if not k.startswith("__")})
    _meta = AttrDict(m)

    return _meta


class DeclarativeFieldsMetaclass(type):
    """
    Metaclass that updates a _meta dict declared on base classes.
    """
    def __new__(mcs, name, bases, attrs):

        # Pop the Meta class if exists
        meta = attrs.pop('Meta', None)

        # Value of abstract by default should be set to false.
        # It is never inherited.
        abstract = getattr(meta, 'abstract', False)

        new_class = super().__new__(mcs, name, bases, attrs)

        _meta = assign_meta(new_class, bases, meta)

        _meta.abstract = abstract

        new_class._meta = _meta

        return new_class



from unittest import TestCase
from canvas_workflow_kit.value_set.hcc2018 import HCCConditions


class TestHCCConditions(TestCase):
    def test_label_hdcc_for(self):
        # code does exist
        result = HCCConditions.label_hdcc_for('M0540')
        expected = 'Rheumatoid Arthritis and Inflammatory Connective Tissue Disease'
        self.assertEqual(expected, result)

        # code does not exist
        result = HCCConditions.label_hdcc_for('XXXX')
        expected = ''
        self.assertEqual(expected, result)

    def test_label_idc10_for(self):
        # code does exist
        result = HCCConditions.label_idc10_for('M0540')
        expected = 'Rheumatoid myopathy with rheumatoid arthritis of unspecified site'
        self.assertEqual(expected, result)

        # code does not exist
        result = HCCConditions.label_idc10_for('XXXX')
        expected = ''
        self.assertEqual(expected, result)

    def test_raf_for(self):
        # code does exist
        result = HCCConditions.raf_for('M0540')
        expected = 0.374
        self.assertEqual(expected, result)

        # code does not exist
        result = HCCConditions.raf_for('XXXX')
        expected = 0
        self.assertEqual(expected, result)



from collections import defaultdict
from typing import Dict


class SuperValueSet(object):
    """
    SuperValueSet allows multiple value-sets to be logically combined
    with |.  So if you need to check if someone has condition A or
    condition B you can:

       patient.has_condition(ConditionA | ConditionB)

    More than two works:

       patient.has_condition(ConditionA | ConditionB | ConditionC)

    """

    def __init__(self, value_sets):
        self.value_sets = value_sets

    @property
    def values(self):
        values: Dict[str, set] = defaultdict(set)

        for value_set in self.value_sets:
            sub_values = value_set.values

            for key in sub_values:
                values[key] |= sub_values[key]

        return values

    @property
    def name(self):
        return ' or '.join([cls.name for cls in self.value_sets])

    def __or__(value_set, other_value_set):
        return SuperValueSet([value_set, other_value_set])


class ValueSystems(type):

    @property
    def values(cls):
        return {
            system.lower(): getattr(cls, system)
            for system in cls.value_systems if hasattr(cls, system)
        }

    @property
    def name(cls):
        return cls.VALUE_SET_NAME

    def __or__(value_set, other_value_set):
        return SuperValueSet([value_set, other_value_set])


class ValueSet(object, metaclass=ValueSystems):

    value_systems = [
        'CANVAS',
        'CPT',
        'CVX',
        'HCPCS',
        'ICD10CM',
        'ICD10PCS',
        'ICD9CM',
        'LOINC',
        'RXNORM',
        'SNOMEDCT',
        'FDB',
        'INTERNAL',
    ]



from typing import Any, Iterable, Optional

import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import InterviewRecommendation
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    BilateralAmputationOfLegBelowOrAboveKnee,
    LeftUnilateralAmputationAboveOrBelowKnee,
    PulseExamOfFoot,
    RightUnilateralAmputationAboveOrBelowKnee,
    SensoryExamOfFoot,
    UnilateralAmputationBelowOrAboveKneeUnspecifiedLaterality,
    VisualExamOfFoot
)
from .diabetes_quality_measure import DiabetesQualityMeasure


class ClinicalQualityMeasure123v6(DiabetesQualityMeasure):
    """
    Diabetes: Foot Exam

    Description: The percentage of patients 18-75 years of age with diabetes (type 1 and type 2)
    who received a foot exam (visual inspection and sensory exam with mono filament and a pulse
    exam) during the measurement year

    Definition: Foot exam: visual inspection with a sensory exam and a pulse exam.

    Rationale: As the seventh leading cause of death in the U.S., diabetes kills approximately
    75,000 people a year (CDC FastStats 2015). Diabetes is a group of diseases marked by high blood
    glucose levels, resulting from the body's inability to produce or use insulin (CDC Statistics
    2014, ADA Basics 2013). People with diabetes are at increased risk of serious health
    complications including vision loss, heart disease, stroke, kidney failure, amputation of toes,
    feet or legs, and premature death. (CDC Fact Sheet 2014).

    In 2012, diabetes cost the U.S. an estimated $245 billion: $176 billion in direct medical costs
    and $69 billion in reduced productivity. This is a 41 percent increase from the estimated $174
    billion spent on diabetes in 2007 (ADA Economic 2013).

    Complications due to improper or poor quality of foot care in diabetics can include amputations
    of the toe, foot, lower and upper leg. Between 1993 and 2009, the CDC monitored the rates of
    nontraumatic amputations in the diabetic population. Overall, the rates peaked in 1996 with toe
    amputations at 3.7 per 1,000 diabetic population. Amputations for the foot, lower leg, and
    upper leg were 1.5, 2.3, and 1.1 per 1,000 diabetic population, respectively, in 1996. Since
    the late nineties, lower extremity amputation rates in the diabetic population have declined,
    witnessed by the following rates in 2009: for toe amputations (1.8 per 1,000 diabetic
    population), followed by lower leg (0.9), foot (0.5) and upper leg (0.4) amputations (CDC,
    2012).

    Guidance: Only patients with a diagnosis of Type 1 or Type 2 diabetes should be included in the
    denominator of this measure; patients with a diagnosis of secondary diabetes due to another
    condition should not be included.

    More information: https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS123v6.html
    """

    class Meta:
        title = 'Diabetes: Foot Exam'

        version = '2019-02-12v1'

        description = (
            'Patients 18-75 years of age with diabetes who have not received a foot exam in the last year.'  # noqa: E501
        )
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS123v6.html'

        identifiers = ['CMS123v6']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'National Committee for Quality Assurance',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
            ClinicalQualityMeasure.CHANGE_PATIENT,
        ]

        references = [
            'American Diabetes Association. Microvascular complications and foot care. Sec. 10. In Standards of Medical Care in Diabetes-2017. Diabetes Care 2017;40(Suppl. 1):S88-S98',  # noqa: E501
            'American Diabetes Association (ADA). 2013. Diabetes Statistics. Retrieved from http://www.diabetes.org/diabetes-basics/diabetes-statistics/.',  # noqa: E501
            'Centers for Disease Control and Prevention (CDC). 2012. CDC\'s Diabetes Program-Data and Trends-Diabetes Surveillance System-Nontraumatic Lower Extremity Amputation with Diabetes by Level of Amputation. Retrieved from http://www.cdc.gov/diabetes/statistics/lealevel/fig8.htm.',  # noqa: E501
            'Centers for Disease Control and Prevention (CDC). 2014. National Diabetes Statistics Report. http://www.cdc.gov/diabetes/pdfs/data/2014-report-estimates-of-diabetes-and-its-burden-in-the-united-states.pdf',  # noqa: E501
            'Centers for Disease Control and Prevention (CDC). 2015. FastStats: Deaths and Mortality. www.cdc.gov/nchs/fastats/deaths.htm.',  # noqa: E501
            'Centers for Disease Control and Prevention. 2014. CDC Features. Diabetes Latest. www.cdc.gov/features/diabetesfactsheet/.',  # noqa: E501
            'American Diabetes Association. 2013. Diabetes Basics. www.diabetes.org/diabetes-basics/?loc=GlobalNavDB',  # noqa: E501
            'American Diabetes Association (ADA). April 2013. Economic Costs of Diabetes in the U.S. in 2012. Diabetes Care. Vol. 36 no. 4 1033-46. http://care.diabetesjournals.org/content/36/4/1033.full',  # noqa: E501
        ]

    @classmethod
    def enabled(cls) -> bool:
        return True

    _on_dates = None

    @property
    def period(self) -> Any:
        if not self._on_dates:
            return 'N/A'
        from_date = min(self._on_dates)
        to_date = max(self._on_dates)
        if from_date == to_date:
            return self.display_date(from_date)
        else:
            return self.display_period(from_date, to_date)

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who have had either a bilateral amputation above or below the knee, or
        both a left and right amputation above or below the knee before or during the measurement
        period.

        Exclude patients who were in hospice care during the measurement year.

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        # bilateral is an exclusion
        conditions = self.patient.conditions.starts_before(self.timeframe.end)

        if conditions.find(BilateralAmputationOfLegBelowOrAboveKnee):
            return False

        # as is two unilateral amputations
        amputations = conditions.find(LeftUnilateralAmputationAboveOrBelowKnee |
                                      RightUnilateralAmputationAboveOrBelowKnee |
                                      UnilateralAmputationBelowOrAboveKneeUnspecifiedLaterality)

        if len(amputations) >= 2:
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Numerator: Patients who received visual, pulse and sensory foot examinations during the
        measurement period

        Exclusions: Not Applicable
        """
        performed = {
            VisualExamOfFoot.OID: False,
            PulseExamOfFoot.OID: False,
            SensoryExamOfFoot.OID: False,
        }
        self._on_dates = []

        interviews = self.patient.interviews.validated().within(self.timeframe)
        for exam_class in [VisualExamOfFoot, PulseExamOfFoot, SensoryExamOfFoot]:
            exams = interviews.find(exam_class)
            if exams and exam_class.OID in performed:
                on_date = max([record['noteTimestamp'] for record in exams])
                self._on_dates.append(arrow.get(on_date))
                performed[exam_class.OID] = True

        return all(performed.values())

    def compute_results(self) -> ProtocolResult:
        """
        Clinical recommendation: American Diabetes Association (2017) Guidelines/ Recommendations:
        Perform annual comprehensive foot examination to identify risk factors predictive of ulcers
        and amputations. The foot examination should include inspection, assessment of foot/leg
        pulses, and testing for loss of protective sensation (10-g monofilament plus testing any
        one of: vibration using 128-Hz tuning fork, pinprick sensation, temperature, ankle
        reflexes, or vibration perception threshold). (Level of evidence: B)
        """
        result = ProtocolResult()

        if self.in_denominator():
            first_name = self.patient.first_name
            if self.in_numerator():
                result.due_in = (
                    min(self._on_dates).shift(days=self.timeframe.duration) - self.now).days
                result.status = STATUS_SATISFIED
                result.add_narrative(
                    f'{first_name} has diabetes and his comprehensive foot exam was done {self.period}.'  # noqa: E501
                )
            else:
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(f'{first_name} has diabetes and is due for foot exam.')
                title = (
                    'Conduct comprehensive foot examination '
                    'including assessment of protective sensation, pulses and visual inspection.')

                # we need to make one interview that satisfies all three of these SNOMED codes
                result.add_interview_recommendation(
                    key='CMS123v6_RECOMMEND_FOOT_EXAM',
                    rank=1,
                    button='Plan',
                    patient=self.patient,
                    questionnaires=[VisualExamOfFoot, PulseExamOfFoot, SensoryExamOfFoot],
                    title=title)
        else:
            result.due_in = self.first_due_in()

        return result
